# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import time
import numpy as np

from cpymad.madx import Madx

import xtrack as xt
import xobjects as xo

# Import thick sequence
mad = Madx()

# CLIC-DR
mad.call('../../test_data/clic_dr/sequence.madx')

# # ELETTRA
# np.sqrt(met[met.loc[:, 'parameter']=='emittance']['mode3'][0]*tw['bets0'])# ELETTRA
# mad.globals.update({'ON_SEXT': 1, 'ON_OCT': 1, 'ON_RF': 1, 'NRJ_GeV': 2.4,
#                    'SAVE_FIGS': False, 'SAVE_TWISS': False})
# mad.call("../../../elettra/elettra2_v15_VADER_2.3T.madx")
# mad.call("../../../elettra/optics_elettra2_v15_VADER_2.3T.madx")

mad.use('ring')

# Twiss
twthick = mad.twiss().dframe()

# Emit
mad.sequence.ring.beam.radiate = True
mad.emit()
mad_emit_table = mad.table.emit.dframe()
mad_emit_summ = mad.table.emitsumm.dframe()

# Makethin
mad.input(f'''
select, flag=MAKETHIN, SLICE=4, thick=false;
select, flag=MAKETHIN, pattern=wig, slice=1;
MAKETHIN, SEQUENCE=ring, MAKEDIPEDGE=true;
use, sequence=RING;
''')
mad.use('ring')
mad.twiss()

# Build xtrack line
print('Build xtrack line...')
line = xt.Line.from_madx_sequence(mad.sequence['RING'])
line.particle_ref = xt.Particles(
        mass0=xt.ELECTRON_MASS_EV,
        q0=-1,
        gamma0=mad.sequence.ring.beam.gamma)

context = xo.ContextCpu()

# Build tracker
print('Build tracker ...')
line.build_tracker(_context=context)
line.matrix_stability_tol = 1e-2

line.configure_radiation(model='mean')

# Twiss
print('Checks with twiss...')
tw = line.twiss(eneloss_and_damping=True)

# Checks
met = mad_emit_table
assert np.isclose(tw['eneloss_turn'], mad_emit_summ.u0[0]*1e9,
                  rtol=3e-3, atol=0)
assert np.isclose(tw['damping_constants_s'][0],
    met[met.loc[:, 'parameter']=='damping_constant']['mode1'][0],
    rtol=3e-3, atol=0
    )
assert np.isclose(tw['damping_constants_s'][1],
    met[met.loc[:, 'parameter']=='damping_constant']['mode2'][0],
    rtol=1e-3, atol=0
    )
assert np.isclose(tw['damping_constants_s'][2],
    met[met.loc[:, 'parameter']=='damping_constant']['mode3'][0],
    rtol=3e-3, atol=0
    )

assert np.isclose(tw['partition_numbers'][0],
    met[met.loc[:, 'parameter']=='damping_partion']['mode1'][0],
    rtol=3e-3, atol=0
    )
assert np.isclose(tw['partition_numbers'][1],
    met[met.loc[:, 'parameter']=='damping_partion']['mode2'][0],
    rtol=1e-3, atol=0
    )
assert np.isclose(tw['partition_numbers'][2],
    met[met.loc[:, 'parameter']=='damping_partion']['mode3'][0],
    rtol=3e-3, atol=0
    )

part_co = tw['particle_on_co']
particles = line.build_particles(
    x_norm=[500., 0, 0], y_norm=[0, 0.0001, 0], zeta=part_co.zeta[0],
    delta=np.array([0,0,1e-2]) + part_co.delta[0],
    nemitt_x=1e-9, nemitt_y=1e-9)

line.configure_radiation(model='quantum')

print('Track 3 particles ...')
num_turns = 5000
t1 = time.time()
line.track(particles, num_turns=num_turns, turn_by_turn_monitor=True)
t2 = time.time()
print(f'Track time: {(t2-t1)/num_turns:.2e} s/turn')
mon = line.record_last_track

import matplotlib.pyplot as plt
plt.close('all')
fig = plt.figure()
ax1 = fig.add_subplot(311)
ax2 = fig.add_subplot(312, sharex=ax1)
ax3 = fig.add_subplot(313, sharex=ax1)

ax1.plot(mon.x[0, :].T)
ax2.plot(mon.y[1, :].T)
ax3.plot(mon.delta[2, :].T)
i_turn = np.arange(num_turns)
ax1.plot(part_co.x[0]
    +(mon.x[0,0]-part_co.x[0])*np.exp(-i_turn*tw['damping_constants_turns'][0]))
ax2.plot(part_co.y[0]
    +(mon.y[1,0]-part_co.y[0])*np.exp(-i_turn*tw['damping_constants_turns'][1]))
ax3.plot(part_co.delta[0]
    +(mon.delta[2,0]-part_co.delta[0])*np.exp(-i_turn*tw['damping_constants_turns'][2]))

plt.show()

# Switch radiation
line.configure_radiation(model='mean')
par_for_emit = line.build_particles(x_norm=50*[0],
                zeta=part_co.zeta[0], delta=part_co.delta[0])
line.configure_radiation(model='quantum')

num_turns=1500
print('Track 50 particles...')
t1 = time.time()
line.track(par_for_emit, num_turns=num_turns, turn_by_turn_monitor=True)
t2 = time.time()
print(f'Track time: {(t2-t1)/num_turns:.2e} s/turn')
mon = line.record_last_track

assert np.isclose(np.std(mon.zeta[:, 750:]),
    np.sqrt(met[met.loc[:, 'parameter']=='emittance']['mode3'][0]*tw['bets0']),
    rtol=0.2, atol=0
    )

assert np.isclose(np.std(mon.x[:, 750:]),
    np.sqrt(met[met.loc[:, 'parameter']=='emittance']['mode1'][0]*tw['betx'][0]),
    rtol=0.2, atol=0
    )

assert np.all(mon.y[:] < 1e-15)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import time
import numpy as np

from cpymad.madx import Madx

import xtrack as xt
import xpart as xp

# Import a thick sequence
mad = Madx()
mad.call('../../test_data/clic_dr/sequence.madx')
mad.use('ring')

# Makethin
mad.input(f'''
select, flag=MAKETHIN, SLICE=4, thick=false;
select, flag=MAKETHIN, pattern=wig, slice=1;
MAKETHIN, SEQUENCE=ring, MAKEDIPEDGE=true;
use, sequence=RING;
''')
mad.use('ring')

# Build xtrack line
print('Build xtrack line...')
line = xt.Line.from_madx_sequence(mad.sequence['RING'])
line.particle_ref = xt.Particles(
        mass0=xt.ELECTRON_MASS_EV,
        q0=-1,
        gamma0=mad.sequence.ring.beam.gamma)

line.build_tracker()

################################
# Enable synchrotron radiation #
################################

# we choose the `mean` mode in which the mean power loss is applied without
# stochastic fluctuations (quantum excitation).
line.configure_radiation(model='mean')

#########
# Twiss #
#########

# In the presence of radiation the stability tolerance needs to be increded to
# allow twiss matrix determinant to be different from one.
line.matrix_stability_tol = 1e-2
tw = line.twiss(eneloss_and_damping=True)

bunch_intensity = 1e11
sigma_z = 5e-3
n_part = int(5e5)
nemitt_x = 0.5e-6
nemitt_y = 0.5e-6

particles = xp.generate_matched_gaussian_bunch(
         num_particles=n_part, total_intensity_particles=bunch_intensity,
         nemitt_x=nemitt_x, nemitt_y=nemitt_y, sigma_z=sigma_z,
         line=line)

assert np.isclose(np.std(particles.x),
                  np.sqrt(tw['dx'][0]**2*np.std(particles.delta)**2
                          + tw['betx'][0]*nemitt_x/mad.sequence.ring.beam.gamma),
                  atol=0, rtol=1e-2)

assert np.isclose(np.std(particles.y),
                  np.sqrt(tw['dy'][0]**2*np.std(particles.delta)**2
                          + tw['bety'][0]*nemitt_y/mad.sequence.ring.beam.gamma),
                  atol=0, rtol=1e-2)

assert np.isclose(np.std(particles.zeta), sigma_z, atol=0, rtol=1e-3)


import xtrack as xt
import numpy as np

bend = xt.Bend(k0=0.4, h=0.3, k1=0.1, length=1)

line = xt.Line(elements=[bend])

line.configure_bend_model(edge='linear', core='expanded')

line.slice_thick_elements(
    slicing_strategies=[xt.Strategy(xt.Teapot(5, mode='thick'))])
line.build_tracker()
line._line_before_slicing.build_tracker()
assert line['e0..3'].parent_name == 'e0'
assert line['e0..3']._parent is line['e0']
assert line['e0..entry_map'].parent_name == 'e0'
assert line['e0..entry_map']._parent is line['e0']
assert line['e0..exit_map'].parent_name == 'e0'
assert line['e0..exit_map']._parent is line['e0']

p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03)
p_ref = p0.copy()
p_slice = p0.copy()

line.track(p_slice)
line._line_before_slicing.track(p_ref)

assert_allclose = np.testing.assert_allclose
assert_allclose = np.testing.assert_allclose
assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

line.to_json('ttt.json')
line2 = xt.Line.from_json('ttt.json')
assert isinstance(line2['e0..3'], xt.ThickSliceBend)
assert line2['e0..3'].parent_name == 'e0'
assert line2['e0..3']._parent is None
assert line2['e0..entry_map'].parent_name == 'e0'
assert line2['e0..entry_map']._parent is None
assert line2['e0..exit_map'].parent_name == 'e0'
assert line2['e0..exit_map']._parent is None

line2.build_tracker()
assert isinstance(line2['e0..3'], xt.ThickSliceBend)
assert line2['e0..3'].parent_name == 'e0'
assert line2['e0..3']._parent is line2['e0']
assert line2['e0..entry_map'].parent_name == 'e0'
assert line2['e0..entry_map']._parent is line2['e0']
assert line2['e0..exit_map'].parent_name == 'e0'
assert line2['e0..exit_map']._parent is line2['e0']

line.track(p_slice, backtrack=True)

assert (p_slice.state == 1).all()
assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import numpy as np

import xtrack as xt

line = xt.Line.from_json('status.json')
line.build_tracker()

line.vars['on_x1'] = 250
assert np.isclose(line.twiss(at_elements=['ip1'])['px'][0], 250e-6,
                  atol=1e-6, rtol=0)

line.vars['on_x1'] = -300
assert np.isclose(line.twiss(at_elements=['ip1'])['px'][0], -300e-6,
                  atol=1e-6, rtol=0)

manager = line._var_management['manager']

manager.find_deps([line.vars['on_x1']])
# returns :
#
# [vars['on_x1'], vars['on_x1hs'], vars['on_dx1hs'], vars['on_x1vl'],
#  vars['on_dx1vl'], vars['on_x1vs'], vars['acbrdv4.r1b1'],
#  mad_elements_dct['mcbrdv.4r1.b1'], mad_elements_dct['mcbrdv.4r1.b1']['kick'],
#  line_dict['mcbrdv.4r1.b1'], line_dict['mcbrdv.4r1.b1'].ksl[0],
#  line_dict['mcbrdv.4r1.b1'].ksl, vars['on_dx1vs'], vars['acbv13.l1b1'],
#  ... ]

lref = line._var_management['lref']

lref['mcbrdv.4r1.b1'].ksl[0].expr
# returns:
# (vars['acbrdv4.r1b1']*vars['bv_aux'])

line.vars['acbrdv4.r1b1'].expr
# returns:
# vars['acbrdv4.r1b1'] = ((((((((4.455799347835793e-07*vars['on_x1vs'])+ ...



import xtrack as xt
import numpy as np

env = xt.load_madx_lattice('../../test_data/sps_thick/sps.seq')
env.vars.load_madx('../../test_data/sps_thick/lhc_q20.str')

# RF set tp stay in the linear region
env['actcse.31632'].voltage = 2500e6
env['actcse.31632'].frequency = 3e6
env['actcse.31632'].lag = 180.

line = env.sps
line.particle_ref = xt.Particles(mass0=xt.ELECTRON_MASS_EV, energy0=10e9)

line.insert('zeta_shift', obj=xt.ZetaShift(), at=0)

tt = line.get_table()

# line.slice_thick_elements(slicing_strategies=[
#     xt.slicing.Strategy(slicing=None), # Default
#     xt.slicing.Strategy(slicing=xt.Teapot(2, mode='thick'), element_type=xt.RBend),
#     xt.slicing.Strategy(slicing=xt.Teapot(8, mode='thick'), element_type=xt.Quadrupole),
# ])

tw4d = line.twiss4d()
tw6d = line.twiss()

line.configure_radiation(model='mean')
line.compensate_radiation_energy_loss()

tw_rad = line.twiss(eneloss_and_damping=True)

# Prepare trim
env['frev0'] = 1. / tw4d.T_rev0
env['circum'] = tw4d.circumference
env['frev_trim'] = 0.

env['zeta_shift'].dzeta = 'circum * frev_trim / frev0'

dfrev = np.linspace(-1, 0.9, 100)
part_x = []
part_y = []
part_zeta = []
damp_cons_x_s = []
demp_const_y_s = []
damp_const_zeta_s = []
eq_gemitt_x = []
eq_gemitt_y = []
eq_gemitt_zeta = []
rad_int_dconst_x_s =[]
delta_ave = []
for dff in dfrev:
    print(f'dfrev: {dff}')
    env['frev_trim'] = dff
    tw = line.twiss(eneloss_and_damping=True,
                    radiation_integrals=True)
    part_x.append(tw.partition_numbers[0])
    part_y.append(tw.partition_numbers[1])
    part_zeta.append(tw.partition_numbers[2])
    eq_gemitt_x.append(tw.eq_gemitt_x)
    eq_gemitt_y.append(tw.eq_gemitt_y)
    eq_gemitt_zeta.append(tw.eq_gemitt_zeta)
    delta_ave.append(tw.delta.mean())

    damp_cons_x_s.append(tw.damping_constants_s[0])
    damp_const_zeta_s.append(tw.damping_constants_s[2])
    demp_const_y_s.append(tw.damping_constants_s[1])

    rad_int_dconst_x_s.append(tw.rad_int_damping_constant_x_s)

# Cast to numpy arrays
part_x = np.array(part_x)
part_y = np.array(part_y)
part_zeta = np.array(part_zeta)
eq_gemitt_x = np.array(eq_gemitt_x)
eq_gemitt_y = np.array(eq_gemitt_y)
eq_gemitt_zeta = np.array(eq_gemitt_zeta)
delta_ave = np.array(delta_ave)
damp_cons_x_s = np.array(damp_cons_x_s)
demp_const_y_s = np.array(demp_const_y_s)
damp_const_zeta_s = np.array(damp_const_zeta_s)
rad_int_dconst_x_s = np.array(rad_int_dconst_x_s)

import matplotlib.pyplot as plt
plt.close('all')

plt.figure(1, figsize=(6.4, 4.8*1.8))
ax1 = plt.subplot(3, 1, 1)
plt.plot(dfrev, delta_ave*1e3)
plt.grid()
plt.ylabel(r'$\delta_\text{ave}$ $[10^{-3}]$')
plt.subplot(3, 1, 2, sharex=ax1)
plt.plot(dfrev, part_x, label='n_x')
plt.plot(dfrev, part_y, label='n_y')
plt.plot(dfrev, part_zeta, label='n_zeta')
plt.plot(dfrev, part_x + part_y + part_zeta, label='total', color='black')
plt.ylabel('Partition numbers')
plt.grid()
plt.legend()
plt.subplot(3, 1, 3, sharex=ax1)
plt.semilogy(dfrev, eq_gemitt_x, label='eq_gemitt_x')
plt.semilogy(dfrev, eq_gemitt_y, label='eq_gemitt_y')
plt.semilogy(dfrev, eq_gemitt_zeta, label='eq_gemitt_zeta')
plt.xlabel(r'$\Delta f_\text{rev}$ [Hz]')
plt.ylim(1e-10, 1e-1)
plt.legend()
plt.grid(True)

plt.show()







import xtrack as xt
import numpy as np

line = xt.Line.from_json(
    '../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
tt = line.get_table()

# Define elements to be used as monitors for orbit correction
# (for LHC all element names starting by "bpm" and not ending by "_entry" or "_exit")
tt_monitors = tt.rows['bpm.*'].rows['.*(?<!_entry)$'].rows['.*(?<!_exit)$']
line.steering_monitors_x = tt_monitors.name
line.steering_monitors_y = tt_monitors.name

# Define elements to be used as correctors for orbit correction
# (for LHC all element namesstarting by "mcb.", containing "h." or "v.")
tt_h_correctors = tt.rows['mcb.*'].rows['.*h\..*']
line.steering_correctors_x = tt_h_correctors.name
tt_v_correctors = tt.rows['mcb.*'].rows['.*v\..*']
line.steering_correctors_y = tt_v_correctors.name

# Reference twiss (no misalignments)
tw_ref = line.twiss4d()

# Introduce misalignments on all quadrupoles
tt = line.get_table()
tt_quad = tt.rows[tt.element_type == 'Quadrupole']
rgen = np.random.RandomState(2) # fix seed for random number generator
shift_x = rgen.randn(len(tt_quad)) * 1e-3 # 1. mm rms shift on all quads
shift_y = rgen.randn(len(tt_quad)) * 1e-3 # 1. mm rms shift on all quads
for nn_quad, sx, sy in zip(tt_quad.name, shift_x, shift_y):
    line[nn_quad].shift_x = sx
    line[nn_quad].shift_y = sy

# Closed twiss fails (closed orbit is not found)
# line.twiss4d()

# Create orbit correction object without running the correction
orbit_correction = line.correct_trajectory(twiss_table=tw_ref, run=False)

# Thread
threader = orbit_correction.thread(ds_thread=500., # correct in sections of 500 m
                                   rcond_long=1e-3)

# prints:
#
# Stop at s=500.0, global rms = [x: 1.11e-03 -> 8.53e-05, y: 2.38e-03 -> 2.64e-04]
# Stop at s=1000.0, global rms = [x: 3.63e-03 -> 1.35e-04, y: 3.79e-03 -> 2.40e-04]
# Stop at s=1500.0, global rms = [x: 1.44e-03 -> 1.22e-04, y: 6.68e-04 -> 1.97e-04]
# Stop at s=2000.0, global rms = [x: 2.88e-03 -> 1.75e-04, y: 1.90e-03 -> 1.96e-04]
# Stop at s=2500.0, global rms = [x: 3.03e-03 -> 2.14e-04, y: 1.22e-03 -> 1.82e-04]
# Stop at s=3000.0, global rms = [x: 2.70e-03 -> 2.19e-04, y: 1.91e-03 -> 1.89e-04]
# Stop at s=3500.0, global rms = [x: 1.12e-02 -> 2.17e-04, y: 2.01e-03 -> 1.66e-04]
# Stop at s=4000.0, global rms = [x: 1.59e-03 -> 2.09e-04, y: 1.16e-03 -> 1.85e-04]
# Stop at s=4500.0, global rms = [x: 2.75e-03 -> 3.16e-04, y: 1.08e-03 -> 1.85e-04]
# Stop at s=5000.0, global rms = [x: 2.55e-03 -> 2.82e-04, y: 4.46e-04 -> 1.75e-04]
# ...

kicks_x_thread = orbit_correction.x_correction.get_kick_values()
kicks_y_thread = orbit_correction.y_correction.get_kick_values()

# Closed twiss after threading (closed orbit is found)
tw_after_thread = line.twiss4d()

# Correct (with custom number of singular values)
orbit_correction.correct(n_singular_values=200)
# prints:
#
# Iteration 0, x_rms: 1.30e-03 -> 2.33e-04, y_rms: 6.91e-04 -> 2.20e-04
# Iteration 1, x_rms: 2.33e-04 -> 2.33e-04, y_rms: 2.20e-04 -> 2.16e-04

# Twiss after closed orbit correction
tw_after = line.twiss4d()

# Extract correction strength
s_x_correctors = orbit_correction.x_correction.s_correctors
s_y_correctors = orbit_correction.y_correction.s_correctors
kicks_x = orbit_correction.x_correction.get_kick_values()
kicks_y = orbit_correction.y_correction.get_kick_values()

#!end-doc-part

# Plots
import matplotlib.pyplot as plt
plt.close('all')

plt.figure(1, figsize=(6.4, 4.8*1.7))
sp1 = plt.subplot(411)
sp1.plot(tw_after_thread.s, tw_after_thread.x * 1e3, label='after thread.')
sp1.plot(tw_after.s, tw_after.x * 1e3, label='after c.o. corr.')
plt.legend(loc='upper right')
plt.ylabel('x [mm]')

sp2 = plt.subplot(412, sharex=sp1)
sp2.stem(s_x_correctors, kicks_x * 1e6)
plt.ylabel(r'x kick [$\mu$rad]')

sp3 = plt.subplot(413, sharex=sp1)
sp3.plot(tw_after_thread.s, tw_after_thread.y * 1e3, label='after thread.')
sp3.plot(tw_after.s, tw_after.y * 1e3)
plt.ylabel('y [mm]')

sp4 = plt.subplot(414, sharex=sp1)
sp4.stem(s_y_correctors, kicks_y * 1e6)
plt.ylabel(r'y kick [$\mu$rad]')
sp4.set_xlabel('s [m]')

plt.subplots_adjust(hspace=0.3, top=0.95, bottom=0.08)

plt.show()


import json
import pathlib

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()


@for_all_test_contexts
def test_freeze_longitudinal_explicit(test_context):

    fname_line = test_data_folder / 'lhc_no_bb/line_and_particle.json'

    # import a line and add reference particle
    with open(fname_line) as fid:
        line_dict = json.load(fid)

    print(f"Test {test_context.__class__}")

    line = xt.Line.from_dict(line_dict['line'])
    line.particle_ref = xp.Particles.from_dict(line_dict['particle'])
    # Build the tracker
    line.build_tracker(_context=test_context)

    # Freeze longitudinal coordinates
    line.freeze_longitudinal()

    # Track some particles with frozen longitudinal coordinates
    particles = line.build_particles(delta=1e-3, x=[-1e-3, 0, 1e-3])
    line.track(particles, num_turns=10)
    particles.move(_context=xo.context_default)
    xo.assert_allclose(particles.delta, 1e-3, rtol=0, atol=1e-12)
    xo.assert_allclose(particles.zeta, 0, rtol=9, atol=1e-12)

    # Twiss with frozen longitudinal coordinates (needs to be 4d)
    twiss = line.twiss(method='4d')
    assert twiss.slip_factor == 0

    # Unfreeze longitudinal coordinates
    line.freeze_longitudinal(False)

    # Track some particles with unfrozen longitudinal coordinates
    particles = line.build_particles(delta=1e-3, x=[-1e-3, 0, 1e-3])
    line.track(particles, num_turns=10)
    particles.move(_context=xo.context_default)
    xo.assert_allclose(particles.delta, 0.00099218, rtol=0, atol=1e-6)

    # Twiss with unfrozen longitudinal coordinates (can be 6d)
    twiss = line.twiss(method='6d')
    xo.assert_allclose(twiss.slip_factor, 0.00032151, rtol=0, atol=1e-6)


@for_all_test_contexts
def test_freeze_longitudinal_context_manager(test_context):

    fname_line = test_data_folder / 'lhc_no_bb/line_and_particle.json'

    # import a line and add reference particle
    with open(fname_line) as fid:
        line_dict = json.load(fid)

    print(f"Test {test_context.__class__}")

    line = xt.Line.from_dict(line_dict['line'])
    line.particle_ref = xp.Particles.from_dict(line_dict['particle'])

    # Build the tracker
    line.build_tracker(_context=test_context)

    with xt.freeze_longitudinal(line):
        # Track some particles with frozen longitudinal coordinates
        particles = line.build_particles(delta=1e-3, x=[-1e-3, 0, 1e-3])
        line.track(particles, num_turns=10)
        particles.move(_context=xo.context_default)
        xo.assert_allclose(particles.delta, 1e-3, rtol=0, atol=1e-12)
        xo.assert_allclose(particles.zeta, 0, rtol=9, atol=1e-12)

        # Twiss with frozen longitudinal coordinates (needs to be 4d)
        twiss = line.twiss(method='4d')
        assert twiss.slip_factor == 0

    # Track some particles with unfrozen longitudinal coordinates
    particles = line.build_particles(delta=1e-3, x=[-1e-3, 0, 1e-3])
    line.track(particles, num_turns=10)
    particles.move(_context=xo.context_default)
    xo.assert_allclose(particles.delta, 0.00099218, rtol=0, atol=1e-6)

    # Twiss with unfrozen longitudinal coordinates (can be 6d)
    twiss = line.twiss(method='6d')
    xo.assert_allclose(twiss.slip_factor, 0.00032151, rtol=0, atol=1e-6)


@for_all_test_contexts
def test_freeze_longitudinal_individual_methods(test_context):

    fname_line = test_data_folder / 'lhc_no_bb/line_and_particle.json'

    # import a line and add reference particle
    with open(fname_line) as fid:
        line_dict = json.load(fid)

    print(f"Test {test_context.__class__}")

    line = xt.Line.from_dict(line_dict['line'])
    line.particle_ref = xp.Particles.from_dict(line_dict['particle'])

    # Build the tracker
    line.build_tracker(_context=test_context)

    # Track some particles with frozen longitudinal coordinates
    particles = line.build_particles(delta=1e-3, x=[-1e-3, 0, 1e-3])
    line.track(particles, num_turns=10, freeze_longitudinal=True)
    particles.move(_context=xo.context_default)
    xo.assert_allclose(particles.delta, 1e-3, rtol=0, atol=1e-12)
    xo.assert_allclose(particles.zeta, 0, rtol=9, atol=1e-12)

    # Twiss with frozen longitudinal coordinates (needs to be 4d)
    twiss = line.twiss(method='4d', freeze_longitudinal=True)
    assert twiss.slip_factor == 0

    # Track some particles with unfrozen longitudinal coordinates
    particles = line.build_particles(delta=1e-3, x=[-1e-3, 0, 1e-3])
    line.track(particles, num_turns=10)
    particles.move(_context=xo.context_default)
    xo.assert_allclose(particles.delta, 0.00099218, rtol=0, atol=1e-6)

    # Twiss with unfrozen longitudinal coordinates (can be 6d)
    twiss = line.twiss(method='6d')
    xo.assert_allclose(twiss.slip_factor, 0.00032151, rtol=0, atol=1e-6)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

# For backwards compatibility
from .masses import PROTON_MASS_EV, ELECTRON_MASS_EV, MUON_MASS_EV, Pb208_MASS_EV, U_MASS_EV



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

# This file will generate and display the following IPAC results:
# https://accelconf.web.cern.ch/ipac2023/doi_per_institute/tupm027/index.html
# Specifications for a new electron cooler of the antiproton decelerator at CERN 


import numpy as np
import xtrack as xt
import xpart as xp
import matplotlib.pyplot as plt

######################################
# lattice parameters for AD at 300 MeV
######################################

#lattice parameters for AD at 300Mev/c
# from https://acc-models.web.cern.ch/acc-models/ad/scenarios/lowenergy/lowenergy.tfs
qx = 5.45020077392
qy = 5.41919929346
dqx=-20.10016919292
dqy=-22.29552755573
circumference = 182.43280000000 #m

# relativistic factors
gamma_rel = 1.04987215550 # at 300 MeV/c

# optics at e-cooler (approximate), in m
beta_x = 10 
beta_y = 4
D_x = 0

# electron cooler parameters
current = 2.4  # A current
length = 1.5 # m cooler length
radius_e_beam = 25*1e-3 #m radius of the electron beam
temp_perp = 100e-3 # <E> [eV] = kb*T
temp_long =  1e-3 # <E> [eV]
magnetic_field = 0.060 # 100 Gauss in ELENA
# idea is to study magnetic field imperfections
magnetic_field_ratio_list = [0,1e-4,5e-4,1e-3] #Iterate over different values of the magnetic field quality to see effect on cooling performance.
#magnetic_field_ratio is the ratio of transverse componenet of magnetic field and the longitudinal component. In the ideal case, the ratio is 0.

# some initial beam parameters
emittance = 35e-6
dp_p = 2e-3 
q0 = 1

# simulation parameters: simulate 20 s of cooling, and take data once every 100 ms
max_time_s = 20
int_time_s = 0.01

# some constants, and simple computations
clight = 299792458.0
mass0 = 938.27208816*1e6 #ev/c^2

beta_rel = np.sqrt(gamma_rel**2 - 1)/gamma_rel
p0c = mass0*beta_rel*gamma_rel #eV/c
T_per_turn = circumference/(clight*beta_rel)

# compute length of simulation, as well as sample interval, in turns
num_turns = int(max_time_s/T_per_turn)
save_interval = int(int_time_s/T_per_turn)

# compute initial beam parameters
x_init = np.sqrt(beta_x*emittance)
y_init = np.sqrt(beta_y*emittance)

particles0 = xp.Particles(
            mass0=mass0,
            p0c=p0c,
            q0=q0,
            x=x_init,
            px=0,
            y=0,
            py=0,
            delta=0,
            zeta=0)        

# arc to do linear tracking
arc = xt.LineSegmentMap(
        qx=qx, qy=qx,
        dqx=dqx, dqy=dqy,
        length=circumference,
        betx=beta_x,
        bety=beta_y,
        dx=D_x)

# compute length of simulation, as well as sample interval, in turns
num_turns = int((max_time_s / T_per_turn).item())
save_interval = int((int_time_s / T_per_turn).item())

# create a monitor object, to reduce holded data
monitor = xt.ParticlesMonitor(start_at_turn=0, stop_at_turn=1,
                        n_repetitions=int(num_turns/save_interval),
                        repetition_period=save_interval,
                        num_particles=1)

##############################
# electron cooler parameters #
##############################  
electron_cooler = xt.ElectronCooler(
                length=length,
                radius_e_beam=radius_e_beam,
                current=current,
                temp_perp=temp_perp,
                temp_long=temp_long,
                magnetic_field=magnetic_field, 
                magnetic_field_ratio=0,
                space_charge_factor=0)

line = xt.Line(elements=[monitor, electron_cooler, arc],element_names=['monitor','electron_cooler','arc'])
line.particle_ref = xp.Particles(mass0=mass0, q0=q0, p0c=p0c)
line.build_tracker()

################
# prepare plot #
################

plt.figure(figsize=(10, 5))
plt.rcParams.update({'font.size': 14}) 

##########################################
# scan over magnetic field imperfections #
##########################################

for magnetic_field_ratio in magnetic_field_ratio_list:        
        
        electron_cooler.magnetic_field_ratio=magnetic_field_ratio
        particles=particles0.copy()
        # track
        line.track(particles, num_turns=num_turns,
              turn_by_turn_monitor=False)
        x = monitor.x[:,:,0]
        px = monitor.px[:,:,0]        
        time = monitor.at_turn[:, 0, 0] * T_per_turn
        # compute action at the end for all turns
        action_x = (x**2/beta_x + beta_x*px**2)
        
        plt.plot(time,action_x*1e6,label='$B_{\\perp}/B_{\\parallel}$='f'{magnetic_field_ratio:.0e}')

plt.xlabel('Time [s]')
plt.ylabel('$J_x$ $[\\mu m]$')
plt.legend()
plt.show()



import json

import numpy as np
import xtrack as xt

with open('../../test_data/hllhc14_no_errors_with_coupling_knobs/line_b1.json',
            'r') as fid:
    dct = json.load(fid)
line = xt.Line.from_dict(dct)

line.build_tracker()

tw_before = line.twiss()

line.match(
    # Portion of the beam line to be modified and initial conditions
    start='mq.33l8.b1',
    end='mq.23l8.b1',
    init=tw_before.get_twiss_init(at_element='mq.33l8.b1'),
    # Dipole corrector strengths to be varied
    vary=[
        xt.Vary(name='acbv30.l8b1', step=1e-10),
        xt.Vary(name='acbv28.l8b1', step=1e-10),
        xt.Vary(name='acbv26.l8b1', step=1e-10),
        xt.Vary(name='acbv24.l8b1', step=1e-10),
    ],
    targets=[
        # I want the vertical orbit to be at 3 mm at mq.28l8.b1 with zero angle
        xt.Target('y', at='mb.b28l8.b1', value=3e-3, tol=1e-4, scale=1),
        xt.Target('py', at='mb.b28l8.b1', value=0, tol=1e-6, scale=1000),
        # I want the bump to be closed
        xt.Target('y', at='mq.23l8.b1', value=tw_before['y', 'mq.23l8.b1'],
                  tol=1e-6, scale=1),
        xt.Target('py', at='mq.23l8.b1', value=tw_before['py', 'mq.23l8.b1'],
                  tol=1e-7, scale=1000),
    ]
)

#!end-doc-part

tw = line.twiss()

import matplotlib.pyplot as plt
plt.close('all')
fig = plt.figure(1, figsize=(6.4*1.2, 4.8*0.8))
ax = fig.add_subplot(111)
ax.plot(tw_before.s, tw_before.y*1000, label='y')
ax.plot(tw.s, tw.y*1000, label='y')
ax.axvline(x=line.get_s_position('mb.b28l8.b1'), color='r', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mcbv.30l8.b1'), color='k', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mcbv.28l8.b1'), color='k', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mcbv.26l8.b1'), color='k', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mcbv.24l8.b1'), color='k', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mq.33l8.b1'), color='g', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mq.23l8.b1'), color='g', linestyle='--', alpha=0.5)
ax.set_xlim(line.get_s_position('mq.33l8.b1') - 10,
            line.get_s_position('mq.23l8.b1') + 10)
ax.set_xlabel('s [m]')
ax.set_ylabel('y [mm]')
ax.set_ylim(-0.5, 10)
plt.subplots_adjust(bottom=.152, top=.9, left=.1, right=.95)
plt.show()

assert np.isclose(tw['y', 'mb.b28l8.b1'], 3e-3, atol=1e-4)
assert np.isclose(tw['py', 'mb.b28l8.b1'], 0, atol=1e-6)
assert np.isclose(tw['y', 'mq.23l8.b1'], tw_before['y', 'mq.23l8.b1'], atol=1e-6)
assert np.isclose(tw['py', 'mq.23l8.b1'], tw_before['py', 'mq.23l8.b1'], atol=1e-7)
assert np.isclose(tw['y', 'mq.33l8.b1'], tw_before['y', 'mq.33l8.b1'], atol=1e-6)
assert np.isclose(tw['py', 'mq.33l8.b1'], tw_before['py', 'mq.33l8.b1'], atol=1e-7)

# There is a bit of leakage in the horizontal plane (due to feed-down from sextupoles)
assert np.isclose(tw['x', 'mb.b28l8.b1'], tw_before['x', 'mb.b28l8.b1'], atol=100e-6)
assert np.isclose(tw['px', 'mb.b28l8.b1'], tw_before['px', 'mb.b28l8.b1'], atol=100e-7)
assert np.isclose(tw['x', 'mq.23l8.b1'], tw_before['x', 'mq.23l8.b1'], atol=100e-6)
assert np.isclose(tw['px', 'mq.23l8.b1'], tw_before['px', 'mq.23l8.b1'], atol=100e-7)
assert np.isclose(tw['x', 'mq.33l8.b1'], tw_before['x', 'mq.33l8.b1'], atol=100e-6)
assert np.isclose(tw['px', 'mq.33l8.b1'], tw_before['px', 'mq.33l8.b1'], atol=100e-7)

# Now I match the bump including the horizontal plane

# I start from scratch
line.vars['acbv30.l8b1'] = 0
line.vars['acbv28.l8b1'] = 0
line.vars['acbv26.l8b1'] = 0
line.vars['acbv24.l8b1'] = 0

line.match(
    start='mq.33l8.b1',
    end='mq.23l8.b1',
    init=tw_before.get_twiss_init(at_element='mq.33l8.b1'),
    vary=[
        xt.Vary(name='acbv30.l8b1', step=1e-10),
        xt.Vary(name='acbv28.l8b1', step=1e-10),
        xt.Vary(name='acbv26.l8b1', step=1e-10),
        xt.Vary(name='acbv24.l8b1', step=1e-10),
        xt.Vary(name='acbh27.l8b1', step=1e-10),
        xt.Vary(name='acbh25.l8b1', step=1e-10),
    ],
    targets=[
        # I want the vertical orbit to be at 3 mm at mq.28l8.b1 with zero angle
        xt.Target('y', at='mb.b28l8.b1', value=3e-3, tol=1e-4, scale=1),
        xt.Target('py', at='mb.b28l8.b1', value=0, tol=1e-6, scale=1000),
        # I want the bump to be closed
        xt.Target('y', at='mq.23l8.b1', value=tw_before['y', 'mq.23l8.b1'],
                  tol=1e-6, scale=1),
        xt.Target('py', at='mq.23l8.b1', value=tw_before['py', 'mq.23l8.b1'],
                  tol=1e-7, scale=1000),
        xt.Target('x', at='mq.23l8.b1', value=tw_before['x', 'mq.23l8.b1'],
                  tol=1e-6, scale=1),
        xt.Target('px', at='mq.23l8.b1', value=tw_before['px', 'mq.23l8.b1'],
                  tol=1e-7, scale=1000),
    ]
)

tw = line.twiss()
assert np.isclose(tw['y', 'mb.b28l8.b1'], 3e-3, atol=1e-4)
assert np.isclose(tw['py', 'mb.b28l8.b1'], 0, atol=1e-6)
assert np.isclose(tw['y', 'mq.23l8.b1'], tw_before['y', 'mq.23l8.b1'], atol=1e-6)
assert np.isclose(tw['py', 'mq.23l8.b1'], tw_before['py', 'mq.23l8.b1'], atol=1e-7)
assert np.isclose(tw['y', 'mq.33l8.b1'], tw_before['y', 'mq.33l8.b1'], atol=1e-6)
assert np.isclose(tw['py', 'mq.33l8.b1'], tw_before['py', 'mq.33l8.b1'], atol=1e-7)

# There is a bit of leakage in the horizontal plane (due to feed-down from sextupoles)
assert np.isclose(tw['x', 'mb.b28l8.b1'], tw_before['x', 'mb.b28l8.b1'], atol=50e-6)
assert np.isclose(tw['px', 'mb.b28l8.b1'], tw_before['px', 'mb.b28l8.b1'], atol=2e-6)
assert np.isclose(tw['x', 'mq.23l8.b1'], tw_before['x', 'mq.23l8.b1'], atol=1e-6)
assert np.isclose(tw['px', 'mq.23l8.b1'], tw_before['px', 'mq.23l8.b1'], atol=1e-7)
assert np.isclose(tw['x', 'mq.33l8.b1'], tw_before['x', 'mq.33l8.b1'], atol=1e-6)
assert np.isclose(tw['px', 'mq.33l8.b1'], tw_before['px', 'mq.33l8.b1'], atol=1e-7)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from .loss_location_refinement import LossLocationRefinement, _skip_in_loss_location_refinement



import numpy as np
from scipy.constants import c as clight

import xtrack as xt
import xpart as xp
import xobjects as xo
from functools import partial

line = xt.Line.from_json(
    '../../test_data/hllhc14_no_errors_with_coupling_knobs/line_b1.json')
# line.cycle('ip1')
line.build_tracker()

for vv in line.vars.get_table().rows[
    'on_x.*|on_sep.*|on_crab.*|on_alice|on_lhcb|corr_.*'].name:
    line.vars[vv] = 0

tw = line.twiss()

df_hz = -50
h_rf = 35640
f_rev = 1/tw.T_rev0
df_rev = df_hz / h_rf
eta = tw.slip_factor
delta_expected = -df_rev / f_rev / eta

line.vars['f_rf'] = 400789598.9858259 + df_hz
tt = line.get_table()
for nn in tt.rows[tt.element_type=='Cavity'].name:
    line.element_refs[nn].absolute_time = 1
    line.element_refs[nn].frequency = line.vars['f_rf']


def merit_function(x, num_turns=10):
    p = line.build_particles(x=x[0], px=x[1], y=x[2], py=x[3], zeta=x[4], delta=x[5])
    line.track(p, num_turns=num_turns, turn_by_turn_monitor=True)
    rec = line.record_last_track
    dx = rec.x[0, :] - rec.x[0, 0]
    dpx = rec.px[0, :] - rec.px[0, 0]
    dy = rec.y[0, :] - rec.y[0, 0]
    dpy = rec.py[0, :] - rec.py[0, 0]
    ddelta = rec.delta[0, :] - rec.delta[0, 0]

    out = np.array(list(dx) + list(dpx) + list(dy) + list(dpy) + list(ddelta))
    # print(x, out)
    return out

num_turns = 20
opt = xt.match.opt_from_callable(partial(merit_function, num_turns=num_turns),
                           x0=np.array(6*[0.]),
                           steps=[1e-9, 1e-10, 1e-9, 1e-10, 1e-4, 1e-7],
                           tar=np.array(5*num_turns*[0.]),
                           tols=np.array(5*num_turns*[1e-10]))
# opt.targets[-1].weight = 1e4
opt.step(5)


x_sol = opt.get_knob_values()
particle_on_co = line.build_particles(
    x=x_sol[0], px=x_sol[1], y=x_sol[2], py=x_sol[3], zeta=x_sol[4],
    delta=x_sol[5])

tw1 = line.twiss(particle_on_co=particle_on_co)

T_rev = tw1.T_rev0 - (tw1.zeta[-1] - tw1.zeta[0])/(tw.beta0*clight)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
ax1 = plt.subplot(2, 1, 1)
plt.plot(tw1.s, tw1.x*1e3, label='x')
plt.ylabel('x [mm]')
plt.grid()
ax2 = plt.subplot(2, 1, 2, sharex=ax1)
plt.plot(tw1.s, tw1.delta*1e3, label='delta')
plt.xlabel('s [m]')
plt.ylabel(r'$\delta$ [$10^{-3}$]')
plt.ylim(-0.5, 0.5)
plt.grid()

plt.suptitle(r'$\Delta f_{\mathrm{rf}}$ = ' +f'{df_hz} Hz, '
             f'expexted $\delta$ = {delta_expected*1e3:.2f}e-3')

plt.show()


# copyright ############################### //
# This file is part of the Xtrack Package.   //
# Copyright (c) CERN, 2023.                 //
# ######################################### //

import xobjects as xo
import xtrack as xt

from ..base_element import BeamElement
from ..general import _pkg_root, _print

import numpy as np


# Random generators need to be a BeamElement to get the LocalParticle API
class RandomUniform(BeamElement):
    _xofields = {
        '_dummy': xo.UInt8,  # TODO: a hack for allocating empty struct on OCL
    }

    allow_track = False

    _extra_c_sources = [
        _pkg_root.joinpath('particles', 'rng_src', 'base_rng.h'),
        _pkg_root.joinpath('random', 'random_src', 'uniform.h')
    ]

    _per_particle_kernels = {
        'sample_uniform': xo.Kernel(
                c_name='RandomUniform_sample',
                args=[
                    xo.Arg(xo.Float64, pointer=True, name='samples'),
                    xo.Arg(xo.Int64, name='n_samples_per_seed')
                ]
            )
        }

    def _sample(self, *args, **kwargs):
        self.sample_uniform(*args, **kwargs)

    def generate(self, n_samples=1000, n_seeds=None, particles=None):
        context = self._context
        if n_seeds is None:
            n_seeds = 1000
        elif particles is not None:
            _print("Warning: both 'particles' and 'n_seeds' are given, but "
                  + "are not compatible. Ignoring 'n_seeds'...")
        n_seeds = int(n_seeds)
        if particles is None:
            particles = xt.Particles(state=np.ones(n_seeds),
                                     x=np.ones(n_seeds), _context=context)
        else:
            particles.move(_context=context)
            n_seeds = len(particles._rng_s1)
        if not particles._has_valid_rng_state():
            particles._init_random_number_generator()

        n_samples_per_seed = int(np.floor(n_samples/n_seeds))
        if n_samples_per_seed < 1:
            raise ValueError("Not enough samples to accomodate all seeds!")

        samples = context.zeros(shape=(n_seeds*n_samples_per_seed,),
                                dtype=np.float64)
        self._sample(particles=particles, samples=samples,
                     n_samples_per_seed=n_samples_per_seed)

        return np.reshape(samples[:n_samples_per_seed*n_seeds],
                    (-1, n_samples_per_seed)
                )

class RandomUniformAccurate(RandomUniform):
    _xofields = {
        '_dummy': xo.UInt8,  # TODO: a hack for allocating an empty struct on OCL
    }

    allow_track = False

    _depends_on = [RandomUniform]

    _extra_c_sources = [
        _pkg_root.joinpath('random','random_src','uniform_accurate.h')
    ]

    _per_particle_kernels = {
        'sample_unif_accuurate': xo.Kernel(
                c_name='RandomUniformAccurate_sample',
                args=[
                    xo.Arg(xo.Float64, pointer=True, name='samples'),
                    xo.Arg(xo.Int64, name='n_samples_per_seed')
                ]
            )
        }

    def _sample(self, *args, **kwargs):
        self.sample_unif_accuurate(*args, **kwargs)


class RandomExponential(RandomUniform):
    _xofields = {
        '_dummy': xo.UInt8,  # TODO: a hack for allocating an empty struct on OCL
    }

    allow_track = False

    _depends_on = [RandomUniform]

    _extra_c_sources = [
        _pkg_root.joinpath('random','random_src','exponential.h')
    ]

    _per_particle_kernels = {
        'sample_exp': xo.Kernel(
                c_name='RandomExponential_sample',
                args=[
                    xo.Arg(xo.Float64, pointer=True, name='samples'),
                    xo.Arg(xo.Int64, name='n_samples_per_seed')
                ]
            )
        }

    def _sample(self, *args, **kwargs):
        self.sample_exp(*args, **kwargs)


class RandomNormal(RandomUniform):
    _xofields = {
        '_dummy': xo.UInt8,  # TODO: a hack for allocating an empty struct on OCL
    }

    allow_track = False

    _depends_on = [RandomUniform]

    _extra_c_sources = [
        _pkg_root.joinpath('random','random_src','normal.h')
    ]

    _per_particle_kernels = {
        'sample_gauss': xo.Kernel(
                c_name='RandomNormal_sample',
                args=[
                    xo.Arg(xo.Float64, pointer=True, name='samples'),
                    xo.Arg(xo.Int64, name='n_samples_per_seed')
                ]
            )
        }

    def _sample(self, *args, **kwargs):
        self.sample_gauss(*args, **kwargs)


class RandomRutherford(RandomUniform):
    _xofields = {
        'lower_val':         xo.Float64,
        'upper_val':         xo.Float64,
        'A':                 xo.Float64,
        'B':                 xo.Float64,
        'Newton_iterations': xo.Int8
    }

    allow_track = False

    _depends_on = [RandomUniform]

    _extra_c_sources = [
        _pkg_root.joinpath('random','random_src','exponential_integral_Ei.h'),
        _pkg_root.joinpath('random','random_src','rutherford.h')
    ]

    _per_particle_kernels = {
        'sample_ruth': xo.Kernel(
                c_name='RandomRutherford_sample',
                args=[
                    xo.Arg(xo.Float64, pointer=True, name='samples'),
                    xo.Arg(xo.Int64, name='n_samples_per_seed')
                ]
            ),
        }

    _kernels = {
        'set_rutherford': xo.Kernel(
                c_name='RandomRutherford_set',
                args=[
                    xo.Arg(xo.ThisClass, name='rng'),
                    xo.Arg(xo.Float64, name='A'),
                    xo.Arg(xo.Float64, name='B'),
                    xo.Arg(xo.Float64, name='lower_val'),
                    xo.Arg(xo.Float64, name='upper_val')
                ]
            )
        }

    def _sample(self, *args, **kwargs):
        self.sample_ruth(*args, **kwargs)

    def __init__(self, **kwargs):
        if '_xobject' not in kwargs:
            kwargs.setdefault('Newton_iterations', 7)
            kwargs.setdefault('lower_val', 1.)
            kwargs.setdefault('upper_val', 1.)
            kwargs.setdefault('A', 0.)
            kwargs.setdefault('B', 0.)

        super().__init__(**kwargs)

        if not isinstance(self._context, xo.ContextCpu):
            raise ValueError('Rutherford random generator is not currently supported on GPU.')

    def set_parameters(self, A, B, lower_val, upper_val):
        self.compile_kernels(particles_class=xt.Particles, only_if_needed=True)
        context = self._buffer.context
        context.kernels.set_rutherford(rng=self, A=A, B=B, lower_val=lower_val, upper_val=upper_val)




# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
from scipy.spatial import ConvexHull

import xobjects as xo
import xtrack as xt

from ..beam_elements import LimitPolygon, XYShift, SRotation, Drift, Marker
from ..line import (Line, _is_thick, _behaves_like_drift, _allow_loss_refinement,
                    _has_backtrack, _is_aperture)

from ..general import _print

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())


def _skip_in_loss_location_refinement(element, line):
    if isinstance(element, xt.Replica):
        return _skip_in_loss_location_refinement(element.resolve(line), line)
    return (hasattr(element, 'skip_in_loss_location_refinement')
            and element.skip_in_loss_location_refinement)

class LossLocationRefinement:

    '''
    Class to refine the location of the lost particles within a line.

    Parameters
    ----------
    line : xtrack.Line
        Line for which the loss location refinement is performed.
    backtrack_line : xtrack.Line (optional)
        Line used to backtrack the lost particles. If None, the backtracking
        line is automatically generated from the line.
    n_theta : int
        Number of angles used to generate the interpolating aperture model.
        If None, the number of angles is automatically determined.
    r_max : float
        Radius larger than the largest aperture radius.
    dr : float
        Radius step used to generate the interpolating aperture model.
    ds : float
        Step in the `s` direction used to generate the interpolating aperture
        model.
    save_refine_lines : bool
        If True, the lines used to refine the loss location are saved.
    allowed_backtrack_types : list
        List of element types through which the backtracking is allowed.
        Elements exposing the attribute `allow_loss_refinement` are automatically
        added to the list.

    '''

    def __init__(self, line, backtrack_line=None,
                 n_theta=None, r_max=None, dr=None, ds=None,
                 save_refine_lines=False,
                 allowed_backtrack_types=[]):

        if backtrack_line is not None:
            raise ValueError('Backtracking line not supported anymore!')

        if line.iscollective:
            self._original_line = line
            self.line = line._get_non_collective_line()
        else:
            self._original_line = line
            self.line = line

        self._context = self.line._context
        assert isinstance(self._context, xo.ContextCpu), (
                "Other contexts are not supported!")

        # Build a polygon and compile the kernel
        temp_poly = LimitPolygon(_buffer=self.line._buffer,
                x_vertices=[1,-1, -1, 1], y_vertices=[1,1,-1,-1])
        na = lambda a : np.array(a, dtype=np.float64)
        temp_poly.impact_point_and_normal(x_in=na([0]), y_in=na([0]),
                                          z_in=na([0]), x_out=na([2]),
                                          y_out=na([2]), z_out=na([0]))

        # Build track kernel with all elements + polygon
        elm_gen = self.line.element_dict.copy()
        elm_gen['_xtrack_temp_poly_'] = temp_poly
        ln_gen = Line(elements=elm_gen,
                      element_names=list(line.element_names) + ['_xtrack_temp_poly_'])
        ln_gen.build_tracker(_buffer=self.line._buffer)
        ln_gen.config.XTRACK_GLOBAL_XY_LIMIT = line.config.XTRACK_GLOBAL_XY_LIMIT
        self._ln_gen = ln_gen

        self.i_apertures, self.apertures = find_apertures(self.line)

        self.save_refine_lines = save_refine_lines
        if save_refine_lines:
            self.refine_lines = {}

        self.n_theta = n_theta
        self.r_max = r_max
        self.dr = dr
        self.ds = ds
        self.allowed_backtrack_types = allowed_backtrack_types

    def refine_loss_location(self, particles, i_apertures=None):

        '''
        Refine the location of the lost particles within the line.

        Parameters
        ----------
        particles : xt.Particles
            Particles for which the loss location is refined.
        i_apertures : list (optional)
            List of indices of the apertures for which the loss location
            is refined. If None, the loss location is refined for all
            apertures.

        '''

        if i_apertures is None:
            i_apertures = self.i_apertures

        for i_ap in i_apertures:
            if np.any((particles.at_element==i_ap) & (particles.state==0)):

                if self.i_apertures.index(i_ap) == 0:
                    logger.warning(
                            'Unable to handle the first aperture in the line')
                    continue

                i_aper_1 = i_ap
                i_aper_0 = self.i_apertures[self.i_apertures.index(i_ap) - 1]
                logger.debug(f'i_aper_1={i_aper_1}, i_aper_0={i_aper_0}')

                s0, s1, _ = generate_interp_aperture_locations(self.line,
                                                   i_aper_0, i_aper_1, self.ds)
                assert s1 >= s0
                if s1 - s0 <= self.ds:
                    logger.debug('s1-s0 < ds: nothing to do')
                    continue

                presence_shifts_rotations = check_for_active_shifts_and_rotations(
                                                    self.line, i_aper_0, i_aper_1)
                logger.debug(f'presence_shifts_rotations={presence_shifts_rotations}')

                if (not(presence_shifts_rotations) and
                   apertures_are_identical(self.line.elements[i_aper_0],
                                           self.line.elements[i_aper_1], self.line)):

                    logger.debug('Replicate mode')
                    (interp_line, i_end_thin_0, i_start_thin_1, s0, s1
                            ) = interp_aperture_replicate(self._context,
                                      self.line,
                                      i_aper_0, i_aper_1,
                                      self.ds,
                                      _ln_gen=self._ln_gen)

                else:

                    logger.debug('Polygon interpolation mode')
                    (interp_line, i_end_thin_0, i_start_thin_1, s0, s1
                            ) = interp_aperture_using_polygons(self._context,
                                      self.line,
                                      i_aper_0, i_aper_1,
                                      self.n_theta, self.r_max, self.dr, self.ds,
                                      _ln_gen=self._ln_gen)

                interp_line._original_line = self._original_line
                part_refine = refine_loss_location_single_aperture(
                        particles, i_aper_1, i_end_thin_0,
                        self.line, interp_line, inplace=True,
                        allowed_backtrack_types=self.allowed_backtrack_types)

                if self.save_refine_lines:
                    interp_line.i_start_thin_0 = i_end_thin_0
                    interp_line.i_start_thin_1 = i_start_thin_1
                    interp_line.s0 = s0
                    interp_line.s1 = s1
                    self.refine_lines[i_ap] = interp_line


def check_for_active_shifts_and_rotations(line, i_aper_0, i_aper_1):

    presence_shifts_rotations = False
    for ii in range(i_aper_0, i_aper_1):
        ee = line.elements[ii]
        if ee.__class__ is SRotation:
            if not np.isclose(ee.angle, 0, rtol=0, atol=1e-15):
                presence_shifts_rotations = True
                break
        if ee.__class__ is XYShift:
            if not np.allclose([ee.dx, ee.dy], 0, rtol=0, atol=1e-15):
                presence_shifts_rotations = True
                break
    return presence_shifts_rotations


def apertures_are_identical(aper1, aper2, line):

    if isinstance(aper1, xt.Replica):
        aper1 = aper1.resolve(line)

    if isinstance(aper2, xt.Replica):
        aper2 = aper2.resolve(line)

    if aper1.__class__ != aper2.__class__:
        return False

    identical = True
    for ff in aper1._fields:
        tt = np.allclose(getattr(aper1, ff), getattr(aper2, ff),
                        rtol=0, atol=1e-15)
        if not tt:
            identical = False
            break
    return identical


def find_apertures(ln_gen):
    i_apertures = []
    apertures = []
    for ii, ee in enumerate(ln_gen.elements):
        if _is_aperture(ee, ln_gen):
            i_apertures.append(ii)
            apertures.append(ee)

    return i_apertures, apertures


def refine_loss_location_single_aperture(particles, i_aper_1, i_end_thin_0,
                    line, interp_line,
                    inplace=True,
                    allowed_backtrack_types=[]):

    mask_part = (particles.state == 0) & (particles.at_element == i_aper_1)

    part_refine = xt.Particles(
                    p0c=particles.p0c[mask_part],
                    mass0=particles.mass0,
                    q0=particles.q0,
                    x=particles.x[mask_part],
                    px=particles.px[mask_part],
                    y=particles.y[mask_part],
                    py=particles.py[mask_part],
                    zeta=particles.zeta[mask_part],
                    delta=particles.delta[mask_part],
                    s=particles.s[mask_part],
                    chi=particles.chi[mask_part],
                    charge_ratio=particles.charge_ratio[mask_part])

    i_start = i_end_thin_0 + 1
    i_stop = i_aper_1

    # Check that we are not backtracking through element types that are not allowed
    for nn in interp_line._original_line.element_names[i_start : i_stop]:
        ee = interp_line._original_line.element_dict[nn]

        can_backtrack = True
        if not _has_backtrack(ee, line):
            can_backtrack = False
        elif not _allow_loss_refinement(ee, line):
            can_backtrack = False

            # Check for override
            if isinstance(ee, xt.Replica):
                ee = ee.resolve(line)
            if isinstance(ee, tuple(allowed_backtrack_types)):
                can_backtrack = True

        if not can_backtrack:
            if _skip_in_loss_location_refinement(ee, line):
                return 'skipped'
            raise TypeError(
                f'Cannot backtrack through element {nn} of type '
                f'{ee.__class__.__name__}')

    with xt.line._preserve_config(line):
        line.config.XTRACK_GLOBAL_XY_LIMIT = None
        line.track(part_refine, ele_start=i_start, ele_stop=i_stop,
                    backtrack='force')

    # Track with extra apertures
    interp_line.track(part_refine)
    # There is a small fraction of particles that are not lost.
    # We verified that they are really at the edge. Their coordinates
    # correspond to the end fo the short line, which is correct

    if inplace:
        indx_sorted = np.argsort(part_refine.particle_id)
        with particles._bypass_linked_vars():
            particles.x[mask_part] = part_refine.x[indx_sorted]
            particles.px[mask_part] = part_refine.px[indx_sorted]
            particles.y[mask_part] = part_refine.y[indx_sorted]
            particles.py[mask_part] = part_refine.py[indx_sorted]
            particles.zeta[mask_part] = part_refine.zeta[indx_sorted]
            particles.s[mask_part] = part_refine.s[indx_sorted]
            particles.delta[mask_part] = part_refine.delta[indx_sorted]
            particles.ptau[mask_part] = part_refine.ptau[indx_sorted]
            particles.rvv[mask_part] = part_refine.rvv[indx_sorted]
            particles.rpp[mask_part] = part_refine.rpp[indx_sorted]
            particles.p0c[mask_part] = part_refine.p0c[indx_sorted]
            particles.gamma0[mask_part] = part_refine.gamma0[indx_sorted]
            particles.beta0[mask_part] = part_refine.beta0[indx_sorted]

    return part_refine

def interp_aperture_replicate(context, line,
                              i_aper_0, i_aper_1,
                              ds, _ln_gen, mode='end',):

    temp_buf = context.new_buffer()

    i_start_thin_1 = find_adjacent_drift(line, i_aper_1, direction='upstream') + 1
    i_end_thin_0 = find_adjacent_drift(line, i_aper_0, direction='downstream') - 1

    s0, s1, s_vect = generate_interp_aperture_locations(line,
                                                   i_aper_0, i_aper_1, ds)

    if mode=='end':
        aper_to_copy = line.elements[i_aper_1]
    elif mode=='start':
        aper_to_copy = line.elements[i_aper_0]
    else:
        raise ValueError(f'Invalid mode: {mode}')
    interp_apertures = []
    for ss in s_vect:
        interp_apertures.append(aper_to_copy.copy(_buffer=temp_buf))

    interp_line = build_interp_line(
            _buffer=temp_buf,
            s0=s0, s1=s1, s_interp=s_vect,
            aper_0=aper_to_copy.copy(_buffer=temp_buf),
            aper_1=aper_to_copy.copy(_buffer=temp_buf),
            aper_interp=interp_apertures,
            line=line, i_start_thin_0=i_end_thin_0,
            i_start_thin_1=i_start_thin_1,
            _ln_gen=_ln_gen)

    return interp_line, i_end_thin_0, i_start_thin_1, s0, s1

def interp_aperture_using_polygons(context, line,
                       i_aper_0, i_aper_1,
                       n_theta, r_max, dr, ds, _ln_gen):

    temp_buf = context.new_buffer()

    polygon_1, i_start_thin_1 = characterize_aperture(line,
                                 i_aper_1, n_theta, r_max, dr,
                                 buffer_for_poly=temp_buf,
                                 coming_from='upstream')

    polygon_0, i_end_thin_0 = characterize_aperture(line, i_aper_0,
                                 n_theta, r_max, dr,
                                 buffer_for_poly=temp_buf,
                                 coming_from='downstream')

    s0, s1, s_vect = generate_interp_aperture_locations(line,
                                                   i_aper_0, i_aper_1, ds)

    Delta_s = s1 - s0
    interp_polygons = []
    for ss in s_vect:
        x_non_convex=(polygon_1.x_vertices*(ss - s0) / Delta_s
                  + polygon_0.x_vertices*(s1 - ss) / Delta_s)
        y_non_convex=(polygon_1.y_vertices*(ss - s0) / Delta_s
                  + polygon_0.y_vertices*(s1 - ss) / Delta_s)
        hull = ConvexHull(np.array([x_non_convex, y_non_convex]).T)
        i_hull = np.sort(hull.vertices)
        x_hull = x_non_convex[i_hull]
        y_hull = y_non_convex[i_hull]
        interp_polygons.append(LimitPolygon(
            _buffer=temp_buf,
            x_vertices=x_hull,
            y_vertices=y_hull))

    interp_line = build_interp_line(
            _buffer=temp_buf,
            s0=s0, s1=s1, s_interp=s_vect,
            aper_0=polygon_0, aper_1=polygon_1,
            aper_interp=interp_polygons,
            line=line, i_start_thin_0=i_end_thin_0,
            i_start_thin_1=i_start_thin_1,
            _ln_gen=_ln_gen)

    return interp_line, i_end_thin_0, i_start_thin_1, s0, s1

def generate_interp_aperture_locations(line, i_aper_0, i_aper_1, ds):

    s0 = line.tracker._tracker_data_base.element_s_locations[i_aper_0]
    s1 = line.tracker._tracker_data_base.element_s_locations[i_aper_1]
    assert s1>=s0
    n_segments = int(np.ceil((s1-s0)/ds))
    if n_segments <= 1:
        s_vect = np.array([])
    else:
        s_vect = np.linspace(s0, s1, n_segments+1)[1:-1]

    return s0, s1, s_vect

def build_interp_line(_buffer, s0, s1, s_interp, aper_0, aper_1, aper_interp,
                         line, i_start_thin_0, i_start_thin_1, _ln_gen):

    # Build interp line
    s_elements = [s0] + list(s_interp) +[s1]
    elements = [aper_0] + aper_interp + [aper_1]

    for i_ele in range(i_start_thin_0+1, i_start_thin_1):
        ee = line.elements[i_ele]
        if not _behaves_like_drift(ee, line):
            assert not _is_thick(ee, line)
            ss_ee = line.tracker._tracker_data_base.element_s_locations[i_ele]
            elements.append(ee.copy(_buffer=_buffer))
            s_elements.append(ss_ee)
    i_sorted = np.argsort(s_elements)
    s_sorted = list(np.take(s_elements, i_sorted))
    ele_sorted = list(np.take(elements, i_sorted))

    s_all = [s_sorted[0]]
    ele_all = [ele_sorted[0]]

    for ii in range(1, len(s_sorted)):
        ss = s_sorted[ii]

        if ss-s_all[-1]>1e-14:
            ele_all.append(Drift(_buffer=_buffer, length=ss-s_all[-1]))
            s_all.append(ss)
        ele_all.append(ele_sorted[ii])
        s_all.append(s_sorted[ii])

    interp_line = Line(elements=ele_all)

    interp_line.build_tracker(_buffer=_buffer,
                              track_kernel=_ln_gen.tracker.track_kernel)
    interp_line.reset_s_at_end_turn = False
    interp_line.config.XTRACK_GLOBAL_XY_LIMIT = _ln_gen.config.XTRACK_GLOBAL_XY_LIMIT


    return interp_line

def find_adjacent_drift(line, i_element, direction):

    ii=i_element
    found = False
    assert direction in ['upstream', 'downstream']
    if direction == 'upstream':
        increment = -1
    else:
        increment = 1
    while not(found):
        ee = line.element_dict[line.element_names[ii]]
        if isinstance(ee, xt.Replica):
            ee = ee.resolve(line)
        ccnn = ee.__class__.__name__
        #_print(ccnn)
        if ccnn.startswith('Drift'):
            found = True
        elif _behaves_like_drift(ee, line):
            found = True
        else:
            ii += increment

    return ii

def find_previous_drift(line, i_aperture):

    ii=i_aperture
    found = False
    while not(found):
        ee = line.element_dict[line.element_names[ii]]
        if isinstance(ee, xt.Replica):
            ee = ee.resolve(line)
        ccnn = ee.__class__.__name__
        if ccnn == 'Drift':
            found = True
        elif _behaves_like_drift(ee, line):
            found = True
        else:
            ii -= 1
    i_start = ii + 1

    return i_start

def index_in_reversed_line(num_elements, ii):
    return num_elements - ii - 1


def characterize_aperture(line, i_aperture, n_theta, r_max, dr,
                          buffer_for_poly, coming_from='upstream'):

    assert coming_from in ['upstream', 'downstream']

    # find previous drift
    if coming_from == 'upstream':
        i_start = find_adjacent_drift(line, i_aperture, 'upstream') + 1
        i_stop = i_aperture + 1
        backtrack = False
        index_start_thin = i_start
    elif coming_from == 'downstream':
        i_stop = find_adjacent_drift(line, i_aperture, 'downstream')
        i_start = i_aperture
        backtrack = 'force'
        assert np.all([_has_backtrack(ee, line) for ee in
                line.tracker._tracker_data_base.elements[i_start:i_stop+1]])
        index_start_thin = i_stop - 1

    # Get polygon
    theta_vect = np.linspace(0, 2*np.pi, n_theta+1)[:-1]

    this_rmin = 0
    this_rmax = r_max
    this_dr = (this_rmax-this_rmin)/100.
    rmin_theta = 0*theta_vect
    t_iter = []
    for iteration in range(2):

        r_vect = np.arange(this_rmin, this_rmax, this_dr)

        RR, TT = np.meshgrid(r_vect, theta_vect)
        RR += np.atleast_2d(rmin_theta).T

        x_test = RR.flatten()*np.cos(TT.flatten())
        y_test = RR.flatten()*np.sin(TT.flatten())

        logger.info(f'iteration={iteration} num_part={x_test.shape[0]}')

        ptest = xt.Particles(p0c=1,
                x = x_test.copy(),
                y = y_test.copy())
        with xt.line._preserve_config(line):
            line.config.XTRACK_GLOBAL_XY_LIMIT = None
            line.track(ptest, ele_start=i_start, ele_stop=i_stop,
                       backtrack=backtrack)

        indx_sorted = np.argsort(ptest.particle_id)
        state_sorted = np.take(ptest.state, indx_sorted)

        state_mat = state_sorted.reshape(RR.shape)

        i_r_aper = np.argmin(state_mat>0, axis=1)

        rmin_theta = r_vect[i_r_aper-1]
        this_rmin=0
        this_rmax=2*this_dr
        this_dr = dr

    x_mat = x_test.reshape(RR.shape)
    y_mat = y_test.reshape(RR.shape)
    x_non_convex = np.array(
            [x_mat[itt, i_r_aper[itt]] for itt in range(n_theta)])
    y_non_convex = np.array(
            [y_mat[itt, i_r_aper[itt]] for itt in range(n_theta)])

    hull = ConvexHull(np.array([x_non_convex, y_non_convex]).T)
    i_hull = np.sort(hull.vertices)
    x_hull = x_non_convex[i_hull]
    y_hull = y_non_convex[i_hull]

    # Get a convex polygon that does not have points for all angles
    temp_poly = LimitPolygon(x_vertices=x_hull, y_vertices=y_hull)

    # Get a convex polygon that has vertices at all requested angles
    r_out = 1. # m
    res = temp_poly.impact_point_and_normal(
            x_in=0*theta_vect, y_in=0*theta_vect, z_in=0*theta_vect,
            x_out=r_out*np.cos(theta_vect),
            y_out=r_out*np.sin(theta_vect),
            z_out=0*theta_vect)

    polygon = LimitPolygon(x_vertices=res[0], y_vertices=res[1],
                              _buffer=buffer_for_poly)

    return polygon, index_start_thin




import json

import numpy as np

import xtrack as xt

###################################
# Load a line and build a tracker #
####################################

with open('../../test_data/hllhc15_noerrors_nobb/line_w_knobs_and_particle.json') as f:
    dct = json.load(f)

line = xt.Line.from_dict(dct['line'])
line.particle_ref = xt.Particles.from_dict(dct['particle'])
line.twiss_default['method'] = '4d'
line.twiss_default['freeze_longitudinal'] = True
line.build_tracker()

vary=[ xt.Vary('ksf.b1', step=1e-8),  xt.Vary('ksd.b1', step=1e-8)]
line.match(
    vary=vary,
    targets = [xt.Target('dqx', 2.0, tol=1e-6),
               xt.Target('dqy', 2.0, tol=1e-6)])

tw0 = line.twiss()
line.match_knob('dqx.b1',
            knob_value_start=2.0,
            knob_value_end=3.0,
            vary=[ xt.Vary('ksf.b1', step=1e-8), xt.Vary('ksd.b1', step=1e-8)],
            targets=[
                xt.Target('dqx', 3.0, tol=1e-6),
                xt.Target('dqy', tw0, tol=1e-6)])

line.match_knob('dqy.b1',
            knob_value_start=2.0,
            knob_value_end=3.0,
            vary=[ xt.Vary('ksf.b1', step=1e-8), xt.Vary('ksd.b1', step=1e-8)],
            targets=[
                xt.Target('dqx', tw0, tol=1e-6),
                xt.Target('dqy', 3.0, tol=1e-6)])

tw = line.twiss()
assert np.isclose(tw.dqx, 2.0, atol=1e-6)
assert np.isclose(tw.dqy, 2.0, atol=1e-6)

line.vars['dqx.b1'] = 6.0
line.vars['dqy.b1'] = 7.0

tw = line.twiss()
assert np.isclose(tw.dqx, 6.0, atol=1e-4)
assert np.isclose(tw.dqy, 7.0, atol=1e-4)











# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import sys
import numpy as np

import xtrack as xt
import xobjects as xo

from cpymad.madx import Madx

# path = '../../test_data/hllhc14_input_mad/'
# mad = Madx(command_log="mad_final.log")
# mad.call(path + "final_seq.madx")
# mad.use(sequence="lhcb1")
# mad.twiss()
# mad.readtable(file=path + "final_errors.tfs", table="errtab")
# mad.seterr(table="errtab")
# mad.set(format=".15g")

mad = Madx()
mad.call('../../test_data/hllhc15_noerrors_nobb/sequence.madx')
mad.use('lhcb1')

# I want only the betatron part in the sigma matrix
mad.sequence.lhcb1.beam.sigt = 1e-10
mad.sequence.lhcb1.beam.sige = 1e-10
mad.sequence.lhcb1.beam.et = 1e-10

twmad = mad.twiss(rmatrix=True, chrom=True)

line = xt.Line.from_madx_sequence(
        mad.sequence['lhcb1'], apply_madx_errors=True)
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1,
                        gamma0=mad.sequence.lhcb1.beam.gamma)

context = xo.ContextCpu()
line.build_tracker(_context=context)

twxt = line.twiss()

import matplotlib.pyplot as plt

plt.close('all')

fig1 = plt.figure(1, figsize=(6.4, 4.8*1.5))
spbet = plt.subplot(3,1,1)
spco = plt.subplot(3,1,2, sharex=spbet)
spdisp = plt.subplot(3,1,3, sharex=spbet)

spbet.plot(twmad['s'], twmad['betx'], 'b', label=r'madx $\beta_x$')
spbet.plot(twxt['s'], twxt['betx'], '--', color='lightblue', label=r'xtrack $\beta_x$')
spbet.plot(twmad['s'], twmad['bety'], 'r', label=r'madx $\beta_y$')
spbet.plot(twxt['s'], twxt['bety'], '--', color='darkred', label=r'xtrack $\beta_y$')
spbet.set_ylabel(r'$\beta_{x,y}$ [m]')
spbet.legend(loc='upper right')

spco.plot(twmad['s'], twmad['x'], 'b')
spco.plot(twxt['s'], twxt['x'], '--', color='lightblue')
spco.plot(twmad['s'], twmad['y'], 'r')
spco.plot(twxt['s'], twxt['y'], '--', color='darkred')
spco.set_ylabel(r'$x,y$ [m]')

spdisp.plot(twmad['s'], twmad['dx'], 'b')
spdisp.plot(twxt['s'], twxt['dx'], '--', color='lightblue')
spdisp.plot(twmad['s'], twmad['dy'], 'r')
spdisp.plot(twxt['s'], twxt['dy'], '--', color='darkred')
spdisp.set_ylabel(r'$D_{x,y}$ [m]')
spdisp.set_xlabel(r'$s$ [m]')

fig1.subplots_adjust(left=.15, right=.92, hspace=.27)

nemitt_x = 2.5e-6
nemitt_y = 2.5e-6
Sigmas = twxt.get_betatron_sigmas(nemitt_x, nemitt_y)

assert np.isclose(mad.table.summ.q1[0], twxt['qx'], rtol=1e-4)
assert np.isclose(mad.table.summ.q2[0], twxt['qy'], rtol=1e-4)
assert np.isclose(mad.table.summ.dq1, twxt['dqx'], atol=0.1, rtol=0)
assert np.isclose(mad.table.summ.dq2, twxt['dqy'], atol=0.1, rtol=0)
assert np.isclose(mad.table.summ.alfa[0],
    twxt['momentum_compaction_factor'], atol=5e-8, rtol=0)
assert np.isclose(twxt['qs'], 0.0021, atol=1e-4, rtol=0)

for name in ['mb.b19r5.b1', 'mb.b19r1.b1', 'ip1', 'ip2', 'ip5', 'ip8',
             'mbxf.4l1', 'mbxf.4l5']:
    imad = list(twmad['name']).index(name+':1')
    ixt = list(twxt['name']).index(name)

    # for sp in [spbet, spco, spdisp]:
    #     sp.axvline(x=twxt['s'][imad])

    assert np.isclose(twxt['betx'][ixt], twmad['betx'][imad],
                      atol=0, rtol=3e-4)
    assert np.isclose(twxt['bety'][ixt], twmad['bety'][imad],
                      atol=0, rtol=3e-4)
    assert np.isclose(twxt['alfx'][ixt], twmad['alfx'][imad],
                      atol=1e-1, rtol=0)
    assert np.isclose(twxt['alfy'][ixt], twmad['alfy'][imad],
                      atol=1e-1, rtol=0)
    assert np.isclose(twxt['dx'][ixt], twmad['dx'][imad], atol=1e-2)
    assert np.isclose(twxt['dy'][ixt], twmad['dy'][imad], atol=1e-2)
    assert np.isclose(twxt['dpx'][ixt], twmad['dpx'][imad], atol=3e-4)
    assert np.isclose(twxt['dpy'][ixt], twmad['dpy'][imad], atol=3e-4)

    assert np.isclose(twxt['s'][ixt], twmad['s'][imad], atol=5e-6)
    assert np.isclose(twxt['x'][ixt], twmad['x'][imad], atol=5e-6)
    assert np.isclose(twxt['y'][ixt], twmad['y'][imad], atol=5e-6)
    assert np.isclose(twxt['px'][ixt], twmad['px'][imad], atol=1e-7)
    assert np.isclose(twxt['py'][ixt], twmad['py'][imad], atol=1e-7)

    assert np.isclose(Sigmas.Sigma11[ixt], twmad['sig11'][imad], atol=5e-10)
    assert np.isclose(Sigmas.Sigma12[ixt], twmad['sig12'][imad], atol=1e-12)
    assert np.isclose(Sigmas.Sigma13[ixt], twmad['sig13'][imad], atol=1e-10)
    assert np.isclose(Sigmas.Sigma14[ixt], twmad['sig14'][imad], atol=1e-12)
    assert np.isclose(Sigmas.Sigma22[ixt], twmad['sig22'][imad], atol=1e-12)
    assert np.isclose(Sigmas.Sigma23[ixt], twmad['sig23'][imad], atol=1e-12)
    assert np.isclose(Sigmas.Sigma24[ixt], twmad['sig24'][imad], atol=1e-12)
    assert np.isclose(Sigmas.Sigma33[ixt], twmad['sig33'][imad], atol=5e-10)
    assert np.isclose(Sigmas.Sigma34[ixt], twmad['sig34'][imad], atol=3e-12)
    assert np.isclose(Sigmas.Sigma44[ixt], twmad['sig44'][imad], atol=1e-12)

    # check matrix is symmetric
    assert np.isclose(Sigmas.Sigma12[ixt], Sigmas.Sigma21[ixt], atol=1e-16)
    assert np.isclose(Sigmas.Sigma13[ixt], Sigmas.Sigma31[ixt], atol=1e-16)
    assert np.isclose(Sigmas.Sigma14[ixt], Sigmas.Sigma41[ixt], atol=1e-16)
    assert np.isclose(Sigmas.Sigma23[ixt], Sigmas.Sigma32[ixt], atol=1e-16)
    assert np.isclose(Sigmas.Sigma24[ixt], Sigmas.Sigma42[ixt], atol=1e-16)
    assert np.isclose(Sigmas.Sigma34[ixt], Sigmas.Sigma43[ixt], atol=1e-16)

    # check matrix consistency with Sigma.Sigma
    assert np.isclose(Sigmas.Sigma11[ixt], Sigmas.Sigma[ixt][0, 0], atol=1e-16)
    assert np.isclose(Sigmas.Sigma12[ixt], Sigmas.Sigma[ixt][0, 1], atol=1e-16)
    assert np.isclose(Sigmas.Sigma13[ixt], Sigmas.Sigma[ixt][0, 2], atol=1e-16)
    assert np.isclose(Sigmas.Sigma14[ixt], Sigmas.Sigma[ixt][0, 3], atol=1e-16)
    assert np.isclose(Sigmas.Sigma21[ixt], Sigmas.Sigma[ixt][1, 0], atol=1e-16)
    assert np.isclose(Sigmas.Sigma22[ixt], Sigmas.Sigma[ixt][1, 1], atol=1e-16)
    assert np.isclose(Sigmas.Sigma23[ixt], Sigmas.Sigma[ixt][1, 2], atol=1e-16)
    assert np.isclose(Sigmas.Sigma24[ixt], Sigmas.Sigma[ixt][1, 3], atol=1e-16)
    assert np.isclose(Sigmas.Sigma31[ixt], Sigmas.Sigma[ixt][2, 0], atol=1e-16)
    assert np.isclose(Sigmas.Sigma32[ixt], Sigmas.Sigma[ixt][2, 1], atol=1e-16)
    assert np.isclose(Sigmas.Sigma33[ixt], Sigmas.Sigma[ixt][2, 2], atol=1e-16)
    assert np.isclose(Sigmas.Sigma34[ixt], Sigmas.Sigma[ixt][2, 3], atol=1e-16)
    assert np.isclose(Sigmas.Sigma41[ixt], Sigmas.Sigma[ixt][3, 0], atol=1e-16)
    assert np.isclose(Sigmas.Sigma42[ixt], Sigmas.Sigma[ixt][3, 1], atol=1e-16)
    assert np.isclose(Sigmas.Sigma43[ixt], Sigmas.Sigma[ixt][3, 2], atol=1e-16)
    assert np.isclose(Sigmas.Sigma44[ixt], Sigmas.Sigma[ixt][3, 3], atol=1e-16)

    # Check sigma_x, sigma_y
    assert np.isclose(Sigmas.sigma_x[ixt], np.sqrt(Sigmas.Sigma11[ixt]), atol=1e-16)
    assert np.isclose(Sigmas.sigma_y[ixt], np.sqrt(Sigmas.Sigma33[ixt]), atol=1e-16)



plt.show()



import numpy as np
from pathlib import Path

from scipy.constants import c as clight
from scipy.constants import e as qe
from scipy.constants import epsilon_0 as eps0

import xobjects as xo
import xtrack as xt
from xtrack._temp.boris_and_solenoid_map.solenoid_field import SolenoidField

ctx = xo.ContextCpu()

# p0 = xt.Particles(mass0=xt.ELECTRON_MASS_EV, q0=1,
#                  energy0=45.6e9,
#                  x=-140e-3, px=10e-3,
#                  y=1e-3, py=0,
#                  delta=0)
p0 = xt.Particles(mass0=xt.ELECTRON_MASS_EV, q0=1,
                 energy0=45.6e9,
                 x=1e-3, px=1e-3,
                 y=1e-3, py=1e-3,
                 delta=0)


p = p0.copy()

z_sol_center = 15
sf = SolenoidField(L=4, a=0.3, B0=2, z0=z_sol_center)

z_axis = np.linspace(0, 30, 200)
Bz_axis = sf.get_field(0 * z_axis, 0 * z_axis, z_axis)[2]

z_fine = np.linspace(0, 30, 1000)
Bz_fine = sf.get_field(0 * z_fine, 0 * z_fine, z_fine)[2]

P0_J = p.p0c[0] * qe / clight
brho = P0_J / qe / p.q0

#ks = 0.5 * (Bz_axis[:-1] + Bz_axis[1:]) / brho
ks = Bz_axis[:-1] / brho

line = xt.Line(elements=[xt.Solenoid(length=z_axis[1]-z_axis[0], ks=ks[ii])
                            for ii in range(len(z_axis)-1)])
line.cut_at_s(np.linspace(0, line.get_length(), 1000))
line.build_tracker()

p_xt = p0.copy()
line.configure_radiation(model=None)
line.track(p_xt, turn_by_turn_monitor='ONE_TURN_EBE')
mon = line.record_last_track

Bz_mid = 0.5 * (Bz_axis[:-1] + Bz_axis[1:])
Bz_mon = 0 * Bz_axis
Bz_mon[1:] = Bz_mid


import matplotlib.pyplot as plt
plt.close('all')

toplot = ['canonical', 'potential', 'prime']
toplot = ['canonical']

i_part_plot = 0
dz = z_axis[1] - z_axis[0]

plt.figure(100, figsize=(6.4, 4.8*1.6))

sp1 = plt.subplot(4, 1, 1)
plt.bar(z_axis - dz/2 - z_sol_center, Bz_axis, width=z_axis[1]-z_axis[0],
        alpha=0.5, align='edge', linewidth=1, edgecolor='C0')
plt.plot(z_fine- z_sol_center, Bz_fine)
plt.ylabel(r'$B_{s}$ [T]')

sp2 = plt.subplot(4, 1, 2, sharex=sp1)
plt.plot(mon.s[i_part_plot, :] - z_sol_center, 1e3 * mon.x[i_part_plot, :], label='x')
plt.plot(mon.s[i_part_plot, :] - z_sol_center, 1e3 * mon.y[i_part_plot, :], label='y')
plt.ylim(-5, 30)
plt.grid()
plt.ylabel('[mm]')
plt.legend(fontsize='medium', loc='upper left')

px = mon.px[i_part_plot, :]
kin_px = mon.kin_px[i_part_plot, :]
kin_xprime = mon.kin_xprime[i_part_plot, :]

py = mon.py[i_part_plot, :]
kin_py = mon.kin_py[i_part_plot, :]
kin_yprime = mon.kin_yprime[i_part_plot, :]

sp3 = plt.subplot(4, 1, 3, sharex=sp1)
if 'canonical' in toplot:
    plt.plot(mon.s[i_part_plot, :] - z_sol_center, 1e6 * (px - px[0]),
             label=r'$p_x$')
if 'potential' in toplot:
    plt.plot(mon.s[i_part_plot, :] - z_sol_center, 1e6 * mon.ax[i_part_plot, :],
             label=r'$a_x$')
if 'prime' in toplot:
    plt.plot(mon.s[i_part_plot, :] - z_sol_center, 1e6 * (kin_xprime - kin_xprime[0]),
            label=r"x'", color='C2', linewidth=2)
plt.grid()
plt.legend(fontsize='medium', loc='upper left')
plt.ylabel(r"[$10^{-6}$]")
plt.ylim(-150, 150)

sp4 = plt.subplot(4, 1, 4, sharex=sp1)
if 'canonical' in toplot:
    plt.plot(mon.s[i_part_plot, :] - z_sol_center, 1e6 * (py - py[0]),
             label=r'$p_y$')
if 'potential' in toplot:
    plt.plot(mon.s[i_part_plot, :] - z_sol_center, 1e6 * mon.ay[i_part_plot, :],
             label=r'$a_y$')
if 'prime' in toplot:
    plt.plot(mon.s[i_part_plot, :] - z_sol_center, 1e6 * (kin_yprime - kin_yprime[0]),
         label=r"$y'$", color='C2', linewidth=2)
plt.grid()
plt.legend(fontsize='medium', loc='upper left')
plt.ylabel(r"[$10^{-6}$]")
plt.ylim(-150, 150)

plt.xlim(-5, 5)
plt.xlabel('s [m]')
plt.subplots_adjust(top=.95, bottom=.06, hspace=.3)

plt.savefig(f'solenoid_tracking_{"_".join(toplot)}.png', dpi=200)


plt.show()




# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xtrack as xt
from pathlib import Path

ctx = xo.ContextCpu()
#ctx = xo.ContextCupy()
#ctx = xo.ContextPyopencl()

part = xt.Particles(_context=ctx, p0c=6.5e12, x=[1,2,3])
part._init_random_number_generator()

generator_to_test = 'RandomUniformAccurate'

class TestElement(xt.BeamElement):
    _xofields={
        'dummy': xo.Float64,
        }

    _depends_on = [xt.RandomUniform, xt.RandomUniformAccurate]

    _extra_c_sources = [
        '''
        /*gpufun*/
        void TestElement_track_local_particle(TestElementData el, LocalParticle* part0){
            //start_per_particle_block (part0->part)
                double rr = !!GENERATOR!!_generate(part);
                LocalParticle_set_x(part, rr);
            //end_per_particle_block
        }
        '''.replace('!!GENERATOR!!', generator_to_test)
        ]

telem = TestElement(_context=ctx)

telem.track(part)

# Use turn-by-turn monitor to acquire some statistics
line=xt.Line(elements=[telem])
line.build_tracker(_buffer=telem._buffer)

line.track(part, num_turns=1e6, turn_by_turn_monitor=True)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure()
for i_part in range(part._capacity):
    x = line.record_last_track.x[i_part, :]
    assert np.all(x>0)
    assert np.all(x<1)
    hstgm, bin_edges = np.histogram(x,  bins=50, range=(0, 1), density=True)

    bin_centers = 0.5*(bin_edges[:-1] + bin_edges[1:])
    plt.plot(bin_centers, hstgm)
    plt.ylim(bottom=0, top=1.1)
    plt.grid(True)
    assert np.allclose(hstgm, 1, rtol=1e-10, atol=0.03)

plt.show()




# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import xtrack as xt

# Load a line and build a tracker
line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_and_particle.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.build_tracker()

# Frequency trim to make
df_hz = 50 # Frequency trim

# Twiss
tw0 = line.twiss()

# Compute corresponding delay to be introduced in the line:
#
# T_rev = h_rf/f_rf
# dt = h_rf/(f_rf + df_hz) - h_rf/f_rf = h_rf/f_rf (1/(1+df_hz/f_rf) - 1)
#                                       ~= h_rf/f_rf * (1 - df_hz/f_rf -1)
#                                       = -h_rf/(f_rf^2) * df_hz
#                                       = -T_rev / f_rf * df_hz
# dzeta = -beta0 * clight * dt = circumference * df_hz / f_rf

h_rf = 35640
f_rf = h_rf/tw0.T_rev0
beta0 = line.particle_ref.beta0[0]
dzeta = tw0.circumference * df_hz / f_rf

# Append delay element to the line
line.unfreeze()
line.append_element(element=xt.ZetaShift(dzeta=dzeta), name='zeta_shift')
line.build_tracker()

# Twiss
tw1 = line.twiss()

# Expected momentum from slip factor (eta = -df_rev / f_rev / delta)
f_rev = 1/tw0.T_rev0
df_rev = df_hz / h_rf
eta = tw0.slip_factor
delta_expected = -df_rev / f_rev / eta

tw1.delta[0]    # is -0.00035789
delta_expected  # is -0.00035803

# Check that particle stays off momentum in multi-turn tracking
p = tw1.particle_on_co.copy()

p_test = line.build_particles(x_norm=0,
            delta=np.linspace(delta_expected - 8e-4, delta_expected + 8e-4, 60))
# p_test = line.build_particles(x_norm=0, delta=delta_expected,
#                               zeta=np.linspace(0, 0.5, 101))

# Track
line.track(p_test, num_turns=1000, turn_by_turn_monitor=True, with_progress=True)
mon = line.record_last_track

# Plot
import matplotlib.pyplot as plt
plt.close('all')
plt.figure()
plt.plot(mon.zeta[:, :].T, mon.delta[:, :].T * 1e4, color='C0')
plt.xlabel(r'$\zeta$ [m]')
plt.ylabel('$\delta$ [$10^{-4}$]')
plt.xlim(-.8, .8)
plt.ylim(delta_expected * 1e4 - 4, delta_expected * 1e4 + 4)
plt.axhline(delta_expected * 1e4, color='r', linestyle='--',
            label='$\delta$ expected')
plt.show()



import xtrack as xt

env = xt.Environment()
env['a'] = 10
env['b'] = 20
env.new('q1', 'Quadrupole', length='2.0 + b', knl=[0, '3*a'])
env.new('q2', 'q1', length='2.0 + b', knl=[0, '3*a'])

assert 'q1' in env.element_dict
assert len(env.ref['a']._find_dependant_targets()) == 7
# is:
# [vars['a'],
#  element_refs['q2'].knl[1],
#  element_refs['q2'],
#  element_refs['q2'].knl,
#  element_refs['q1'],
#  element_refs['q1'].knl[1],
#  element_refs['q1'].knl]

env._remove_element('q1')
assert 'q1' not in env.element_dict
assert len(env.ref['a']._find_dependant_targets()) == 4
# is:
# [vars['a'],
#  element_refs['q2'].knl[1],
#  element_refs['q2'],
#  element_refs['q2'].knl]



import xtrack as xt

from cpymad.madx import Madx

mad = Madx()

#########################################
# Laod MAD-X model and import in Xsuite #
#########################################

mad = Madx()
mad.call('lep.seq9')
mad.call('lep.opt9')
mad.input('beam, particle=positron, energy=100, radiate=true;')
mad.use(sequence='lep')
mad.input('vrfc:=2.; vrfsc:=50.; vrfscn:=50.; ! LEP2  rf on')

line = xt.Line.from_madx_sequence(
    sequence=mad.sequence.LEP,
    allow_thick=True,
    enable_align_errors=True,
    deferred_expressions=True,
)
line.particle_ref = xt.Particles(mass0=xt.ELECTRON_MASS_EV,
                            gamma0=mad.sequence.LEP.beam.gamma)

#####################
# Slice Xsuite line #
#####################

line_thick = line.select() # Shallow copy
line.slice_thick_elements(
    slicing_strategies=[
        xt.Strategy(slicing=xt.Teapot(1)),
        xt.Strategy(slicing=xt.Teapot(4), element_type=xt.Bend),
        xt.Strategy(slicing=xt.Teapot(4), element_type=xt.Quadrupole),
    ])
line.to_json('lep_thin.json')

##################################
# Compute equilibrium emittances #
##################################

tw_norad = line.twiss()
line.configure_radiation('mean')
tw_rad = line.twiss(eneloss_and_damping=True)

##############################
# Compare against MAD-X emit #
##############################

mad.input('''
    beam, sequence=lep, particle=POSITRON, pc=100, radiate=true;
    use, sequence=lep;
    emit;
    ''')

mad_eneloss_turn = mad.table.emitsumm.u0[0]
mad_gemitt_x = mad.table.emitsumm.ex[0]
mad_gemitt_zeta = mad.table.emitsumm.et[0]
met = mad.table.emit.dframe()
mad_damping_constant_s_x = met[met.loc[:, 'parameter']=='damping_constant']['mode1'].values[0]
mad_damping_constant_s_y = met[met.loc[:, 'parameter']=='damping_constant']['mode2'].values[0]
mad_damping_constant_s_z = met[met.loc[:, 'parameter']=='damping_constant']['mode3'].values[0]
mad_partition_x = met[met.loc[:, 'parameter']=='damping_partion']['mode1'].values[0]
mad_partition_y = met[met.loc[:, 'parameter']=='damping_partion']['mode2'].values[0]
mad_partition_z = met[met.loc[:, 'parameter']=='damping_partion']['mode3'].values[0]

print(f'Energy loss per turn [eV]:')
print(f'- MAD-X:  {mad_eneloss_turn*1e9:.6g}')
print(f'- Xsuite: {tw_rad.eneloss_turn:.6g}')
print(f'Geometric emittance in x [m.rad]:')
print(f'- MAD-X:  {mad_gemitt_x:.6g}')
print(f'- Xsuite: {tw_rad.eq_gemitt_x:.6g}')
print(f'Geometric emittance in zeta [m.rad]:')
print(f'- MAD-X:  {mad_gemitt_zeta:.6g}')
print(f'- Xsuite: {tw_rad.eq_gemitt_zeta:.6g}')
print('Damping constants x [1/s]:')
print(f'- MAD-X:  {mad_damping_constant_s_x:.6g}')
print(f'- Xsuite: {tw_rad.damping_constants_s[0]:.6g}')
print('Damping constants y [1/s]:')
print(f'- MAD-X:  {mad_damping_constant_s_y:.6g}')
print(f'- Xsuite: {tw_rad.damping_constants_s[1]:.6g}')
print('Damping constants z [1/s]:')
print(f'- MAD-X:  {mad_damping_constant_s_z:.6g}')
print(f'- Xsuite: {tw_rad.damping_constants_s[2]:.6g}')

###########
# Asserts #
###########

import xobjects as xo
xo.assert_allclose(tw_rad['eneloss_turn'],  mad.table.emitsumm.u0*1e9, atol=0, rtol=1e-5)
xo.assert_allclose(tw_rad['eq_gemitt_x'],  mad.table.emitsumm.ex, atol=0, rtol=5e-3)
xo.assert_allclose(tw_rad['eq_gemitt_zeta'],  mad.table.emitsumm.et, atol=0, rtol=1e-3)

xo.assert_allclose(tw_rad['damping_constants_s'][0],
    mad_damping_constant_s_x,
    rtol=3e-3, atol=0
    )
xo.assert_allclose(tw_rad['damping_constants_s'][1],
    mad_damping_constant_s_y,
    rtol=1e-3, atol=0
    )
xo.assert_allclose(tw_rad['damping_constants_s'][2],
    mad_damping_constant_s_z,
    rtol=3e-3, atol=0
    )

xo.assert_allclose(tw_rad['partition_numbers'][0],
    mad_partition_x,
    rtol=3e-3, atol=0
    )
xo.assert_allclose(tw_rad['partition_numbers'][1],
    mad_partition_y,
    rtol=1e-3, atol=0
    )
xo.assert_allclose(tw_rad['partition_numbers'][2],
    mad_partition_z,
    rtol=3e-3, atol=0
    )



import xtrack as xt
import numpy as np

line = xt.Line.from_json(
    '../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
tt = line.get_table()

# Define elements to be used as monitors for orbit correction
# (for LHC all element names starting by "bpm" and not ending by "_entry" or "_exit")
tt_monitors = tt.rows['bpm.*'].rows['.*(?<!_entry)$'].rows['.*(?<!_exit)$']
line.steering_monitors_x = tt_monitors.name
line.steering_monitors_y = tt_monitors.name

# Define elements to be used as correctors for orbit correction
# (for LHC all element namesstarting by "mcb.", containing "h." or "v.")
tt_h_correctors = tt.rows['mcb.*'].rows['.*h\..*']
line.steering_correctors_x = tt_h_correctors.name
tt_v_correctors = tt.rows['mcb.*'].rows['.*v\..*']
line.steering_correctors_y = tt_v_correctors.name

# Reference twiss (no misalignments)
tw_ref = line.twiss4d()

# Introduce misalignments on all quadrupoles
tt = line.get_table()
tt_quad = tt.rows[tt.element_type == 'Quadrupole']
rgen = np.random.RandomState(1) # fix seed for random number generator
shift_x = rgen.randn(len(tt_quad)) * 0.01e-3 # 0.01 mm rms shift on all quads
shift_y = rgen.randn(len(tt_quad)) * 0.01e-3 # 0.01 mm rms shift on all quads
for nn_quad, sx, sy in zip(tt_quad.name, shift_x, shift_y):
    line[nn_quad].shift_x = sx
    line[nn_quad].shift_y = sy

# Twiss before correction
tw_before = line.twiss4d()

# Correct orbit using 5 correctors in each plane
orbit_correction = line.correct_trajectory(twiss_table=tw_ref, n_micado=5,
                                           n_iter=1)
# prints:
#
# Iteration 0, x_rms: 1.65e-03 -> 1.06e-04, y_rms: 2.25e-03 -> 1.76e-04

# Twiss after correction
tw_after = line.twiss4d()

# Extract correction strength
s_x_correctors = orbit_correction.x_correction.s_correctors
s_y_correctors = orbit_correction.y_correction.s_correctors
kicks_x = orbit_correction.x_correction.get_kick_values()
kicks_y = orbit_correction.y_correction.get_kick_values()

#!end-doc-part

# Plots
import matplotlib.pyplot as plt
plt.close('all')

plt.figure(1, figsize=(6.4, 4.8*1.7))
sp1 = plt.subplot(411)
sp1.plot(tw_before.s, tw_before.x * 1e3,
        label=f'before corr. (rms: {tw_before.x.std() * 1e3:.2e} mm)')
sp1.plot(tw_after.s, tw_after.x * 1e3,
        label=f'after corr. (rms: {tw_after.x.std() * 1e3:.2e} mm)')
plt.legend(loc='lower right')
plt.ylabel('x [mm]')

sp2 = plt.subplot(412, sharex=sp1)
sp2.stem(s_x_correctors, kicks_x * 1e6)
plt.ylabel(r'x kick [$\mu$rad]')

sp3 = plt.subplot(413, sharex=sp1)
sp3.plot(tw_before.s, tw_before.y * 1e3,
        label=f'before corr. (rms: {tw_before.y.std() * 1e3:.2e} mm)')
sp3.plot(tw_after.s, tw_after.y * 1e3,
        label=f'after corr. (rms: {tw_after.y.std() * 1e3:.2e} mm)')
plt.ylabel('y [mm]')
plt.legend(loc='lower right')

sp4 = plt.subplot(414, sharex=sp1)
sp4.stem(s_y_correctors, kicks_y * 1e6)
plt.ylabel(r'y kick [$\mu$rad]')
sp4.set_xlabel('s [m]')

plt.subplots_adjust(hspace=0.3, top=0.95, bottom=0.08)

plt.show()


import numpy as np

import xtrack as xt

import matplotlib.pyplot as plt

nemitt_x = 1e-6
nemitt_y = 1e-6


line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, p0c=7e12)
line.build_tracker()

plt.close('all')
plt.figure(1)

# Compute and plot footprint
fp0 = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y)
fp0.plot(color='k', label='I_oct=0')

# Change octupoles strength and compute footprint again
line.vars['i_oct_b1'] = 500
fp1 = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y)
fp1.plot(color='r', label='I_oct=500')

# Change octupoles strength and compute footprint again
line.vars['i_oct_b1'] = -250
fp2 = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y)
fp2.plot(color='b', label='I_oct=-250')

plt.legend()

plt.figure(2)

line.vars['i_oct_b1'] = 0
fp0_jgrid = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                         mode='uniform_action_grid')
fp0_jgrid.plot(color='k', label='I_oct=0')

# Change octupoles strength and compute footprint again
line.vars['i_oct_b1'] = 500
fp1_jgrid = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                            mode='uniform_action_grid')
fp1_jgrid.plot(color='r', label='I_oct=500')

# Change octupoles strength and compute footprint again
line.vars['i_oct_b1'] = -250
fp2_jgrid = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                         mode='uniform_action_grid')
fp2_jgrid.plot(color='b', label='I_oct=-250')

plt.legend()
plt.show()

#!end-doc-part

assert hasattr(fp0, 'theta_grid')
assert hasattr(fp0, 'r_grid')

assert np.isclose(fp0.r_grid[0], 0.1, rtol=0, atol=1e-10)
assert np.isclose(fp0.r_grid[-1], 6, rtol=0, atol=1e-10)

assert np.isclose(fp0.theta_grid[0], 0.05, rtol=0, atol=1e-10)
assert np.isclose(fp0.theta_grid[-1], np.pi/2-0.05, rtol=0, atol=1e-10)





import xtrack as xt

env = xt.Environment()

# We can associate a reference particle to the environment
# which will be passed to all lines generated from this environment
env.partice_ref = xt.Particles(p0c=2e9, mass0=xt.PROTON_MASS_EV)

#############
# Variables #
#############

# The environment can be used to create and inspect variables and associated
# deferred expressions
env['a'] = 3.
env['b'] = '2 * a'

env['b'] # is 6.0
env.get_expr('b') # is (2.0 * vars['a'])
env.info('b')
# prints:
#  vars['b']._get_value()
#  vars['b'] = 6.0

#  vars['b']._expr
#  vars['b'] = (2.0 * vars['a'])

#  vars['b']._expr._get_dependencies()
#  vars['a'] = 3.0

#  vars['b'] does not influence any target

# We can use the eval method to evaluate the value for a given expression:
env.eval('sqrt(b + a)') # returns 3.0

# We can inspect the value and expression of all variables in the environment
env.vars.get_table()
# returns:
#
# Table: 3 rows, 3 cols
# name             value expr
# t_turn_s             0 None
# a                    3 None
# b                    6 (2.0 * vars['a'])

############
# Elements #
############

# The environment can be used to create and inspect elements
env.new('mq', xt.Quadrupole, length='5*a', k1='b')

env['mq'] # accesses the quadrupole

env.info('mq')
# prints:
# Element of type:  Quadrupole
# k1                  6.0                              vars['b']
# k1s                 0.0                              None
# length              15.0                             (5.0 * vars['a'])
# num_multipole_kicks 0                                None
# order               5                                None
# inv_factorial_order 0.008333333333333333             None
# knl                 [0. 0. 0. 0. 0. 0.]              None
# ...

env['mq'].length # is 15.0
env['mq'].k1 # is 6.0

env['mq'].get_expr('length') # is (5.0 * vars['a'])
env['mq'].get_info('length')
# prints:
#  element_refs['mq'].length._get_value()
#  element_refs['mq'].length = 15.0
#
#  element_refs['mq'].length._expr
#  element_refs['mq'].length = (5.0 * vars['a'])
#
#  element_refs['mq'].length._expr._get_dependencies()
#  vars['a'] = 3.0
#
#  element_refs['mq'].length does not influence any target

# It is also possible to add to the environment elements that are instantiated by
# the user:
mysext = xt.Sextupole(length=1.0, k2=0.1)
env.elements['ms'] = mysext

env['ms'] # accesses the sextupole
env.info('ms')

# prints:
# Element of type:  Sextupole
# k2                  0.1                                     None
# k2s                 0.0                                     None
# length              1.0                                     None
# order               5                                       None
# inv_factorial_order 0.008333333333333333                    None
# knl                 [0. 0. 0. 0. 0. 0.]                     None
# ...

#########
# Lines #
#########

# The environment can be used to create and access lines

line1 = env.new_line(
    name='l1', components=['mq', env.new('dd', xt.Drift, length=10)])

line2 = env.new_line(
    name='l2', components=[
        env.place('dd'), # place already created elements
        env.new('ip', xt.Marker)])

env['l1'] # accesses the first line
env['l2'] # accesses the second line

# The environment is associated to the line
line1.env # is env
line2.env # is env

# The variables and elements of the environment are shared by all the lines
# created by the environment
line2['a'] # is 3.0
line2['a'] = 4.0
env['a'] # is 4.0

# All methods available on the environment are also available on the lines, e.g.:
line2.info('a') # is equivalent to env.info('a')



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xtrack as xt

context = xo.ContextCpu()

np2ctx = context.nparray_to_context_array
ctx2np = context.nparray_from_context_array


x_vertices=np.array([1.5, 0.2, -1, -1,  1])*1e-2
y_vertices=np.array([1.3, 0.5,  1, -1, -1])*1e-2

aper = xt.LimitPolygon(
                _context=context,
                x_vertices=np2ctx(x_vertices),
                y_vertices=np2ctx(y_vertices))

# Try some particles inside
parttest = xt.Particles(
                _context=context,
                p0c=6500e9,
                x=x_vertices*0.99,
                y=y_vertices*0.99)
aper.track(parttest)
assert np.allclose(ctx2np(parttest.state), 1)

# Try some particles outside
parttest = xt.Particles(
                _context=context,
                p0c=6500e9,
                x=x_vertices*1.01,
                y=y_vertices*1.01)
aper.track(parttest)
assert np.allclose(ctx2np(parttest.state), 0)

part_gen_range = 0.02
n_part=10000
particles = xt.Particles(
                _context=context,
                p0c=6500e9,
                x=np.random.uniform(-part_gen_range, part_gen_range, n_part),
                px = np.zeros(n_part),
                y=np.random.uniform(-part_gen_range, part_gen_range, n_part),
                py = np.zeros(n_part),
                zeta = np.zeros(n_part),
                delta = np.zeros(n_part))

aper.track(particles)

part_id = context.nparray_from_context_array(particles.particle_id)
part_state = context.nparray_from_context_array(particles.state)
part_x = context.nparray_from_context_array(particles.x)
part_y = context.nparray_from_context_array(particles.y)

import matplotlib.pyplot as plt
plt.close('all')
fig1 = plt.figure(1)
ax = fig1.add_subplot(111)
plt.plot(part_x, part_y, '.', color='red')
plt.plot(part_x[part_state>0], part_y[part_state>0], '.', color='green')
ax.plot(aper.x_closed, aper.y_closed, linewidth=3, color='black')

plt.show()



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xtrack as xt

collider = xt.Environment.from_json(
    '../../test_data/hllhc15_thick/hllhc15_collider_thick.json')
collider.build_trackers()

collider.lhcb1.twiss_default['method'] = '4d'
collider.lhcb2.twiss_default['method'] = '4d'
collider.lhcb2.twiss_default['reverse'] = True

# Twiss B1 and B2 from IP1
two = collider.twiss(start='ip7', end='ip3', init_at='ip1',
                 betx=.15, bety=.15)

# Twiss B1 only from IP1
two1 = collider.lhcb1.twiss(start='ip7', end='ip3', init_at='ip1',
                            betx=.15, bety=.15)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.plot(two.lhcb1.s, two.lhcb1.betx, label='betx')
plt.plot(two.lhcb2.s, two.lhcb2.betx, label='betx')
plt.show()


import xtrack as xt

# Load a line and build a tracker
line = xt.Line.from_json('../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line.build_tracker()

# Match tunes and chromaticities to assigned values
opt = line.match(
    method='4d', # <- passed to twiss
    vary=[
        xt.VaryList(['kqtf.b1', 'kqtd.b1'], step=1e-8, tag='quad'),
        xt.VaryList(['ksf.b1', 'ksd.b1'], step=1e-4, limits=[-0.1, 0.1], tag='sext'),
    ],
    targets = [
        xt.TargetSet(qx=62.315, qy=60.325, tol=1e-6, tag='tune'),
        xt.TargetSet(dqx=10.0, dqy=12.0, tol=0.01, tag='chrom'),
    ])

# Inspect optimization log
opt.log()
# prints:
#
# Table: 3 rows, 18 cols
# iteration     penalty alpha tag tol_met target_active hit_limits vary_active      vary_0 ...
#         0     12.9073    -1     nnnn    yyyy          nnnn       yyyy                  0
#         1  0.00270443     0     nnyy    yyyy          nnnn       yyyy         4.2729e-05
#         2 1.22005e-06     0     yyyy    yyyy          nnnn       yyyy        4.27163e-05

# Inspect optimization outcome
opt.target_status()
opt.vary_status()
# prints:
#
# Target status:
# id state tag   tol_met      residue current_val target_val description
#  0 ON    tune     True  8.53717e-11      62.315     62.315 'qx', val=62.315, tol=1e-06, weight=10)
#  1 ON    tune     True  1.49214e-13      60.325     60.325 'qy', val=60.325, tol=1e-06, weight=10)
#  2 ON    chrom    True  1.22005e-06          10         10 'dqx', val=10, tol=0.01, weight=1)
#  3 ON    chrom    True -1.87538e-09          12         12 'dqy', val=12, tol=0.01, weight=1)
# Vary status:
# id state tag  name    lower_limit  current_val upper_limit val_at_iter_0   step weight
#  0 ON    quad kqtf.b1        None  4.27163e-05        None             0  1e-08      1
#  1 ON    quad kqtd.b1        None -4.27199e-05        None             0  1e-08      1
#  2 ON    sext ksf.b1         -0.1    0.0118965         0.1             0 0.0001      1
#  3 ON    sext ksd.b1         -0.1   -0.0232137         0.1             0 0.0001      1

# Get knob values after optimization
knobs_after_match = opt.get_knob_values()
# contains: {'kqtf.b1': 4.27163e-05,  'kqtd.b1': -4.27199e-05,
#            'ksf.b1': 0.0118965, 'ksd.b1': -0.0232137}

# Get knob values before optimization
knobs_before_match = opt.get_knob_values(iteration=0)
# contains: {'kqtf.b1': 0, 'kqtd.b1': 0, 'ksf.b1': 0, 'ksd.b1': 0}


import time
import multiprocessing as mp
import pickle

import numpy as np

import xtrack as xt
import xdeps as xd

# xt._print.suppress = True



class ActionArcPhaseAdvanceFromCell(xd.Action):
    def __init__(self, arc_name, line_name, line):
        assert arc_name in ['12', '23', '34', '45', '56', '67', '78', '81']
        assert line_name in ['lhcb1', 'lhcb2']
        self.arc_name = arc_name
        self.line_name = line_name
        self.line = line

        beam_number = line_name[-1:]
        sector_start_number = arc_name[:1]
        sector_end_number = arc_name[1:]
        self.start_cell = f's.cell.{arc_name}.b{beam_number}'
        self.end_cell = f'e.cell.{arc_name}.b{beam_number}'
        self.start_arc = f'e.ds.r{sector_start_number}.b{beam_number}'
        self.end_arc = f's.ds.l{sector_end_number}.b{beam_number}'

    def run(self):

        if isinstance(self.line, xt.Environment):
            line = self.line[self.line_name]

        twinit_cell = line.twiss(
                    start=self.start_cell, end=self.end_cell,
                    init='periodic', only_twiss_init=True)
        #  twinit_cell.element_name is start_cell for b1 and end_cell for b2

        tw_to_end_arc = line.twiss(init=twinit_cell,
            start=twinit_cell.element_name, end=self.end_arc)
        tw_to_start_arc = line.twiss(init=twinit_cell,
            start=self.start_arc,end=twinit_cell.element_name)

        mux_arc_from_cell = (tw_to_end_arc['mux', self.end_arc]
                             - tw_to_start_arc['mux', self.start_arc])
        muy_arc_from_cell = (tw_to_end_arc['muy', self.end_arc]
                             - tw_to_start_arc['muy', self.start_arc])

        return {
            'mux_arc_from_cell': mux_arc_from_cell,
            'muy_arc_from_cell': muy_arc_from_cell,
            'twinit_cell': twinit_cell,
            'tw_to_end_arc': tw_to_end_arc,
            'tw_to_start_arc': tw_to_start_arc}

class ActionMatchPhaseWithMQTs(xd.Action):

    def __init__(self, arc_name, line_name, line,
                 mux_arc_target, muy_arc_target, restore=False):

        self.action_arc_phase = ActionArcPhaseAdvanceFromCell(
            arc_name=arc_name, line_name=line_name, line=line)
        self.line = line
        self.mux_arc_target = mux_arc_target
        self.muy_arc_target = muy_arc_target
        self.restore = restore

        beam_number = line_name[-1:]
        self.mqt_knob_names = [
            f'kqtf.a{arc_name}b{beam_number}',
            f'kqtd.a{arc_name}b{beam_number}']

    def run(self):
        #store initial knob values
        mqt_knob_values = {
            kk: self.line.vars[kk]._value for kk in self.mqt_knob_names}

        opt = xd.Optimize(
            targets=[
                xd.Target(action=self.action_arc_phase, tar='mux_arc_from_cell',
                            value=self.mux_arc_target, tol=1e-8),
                xd.Target(action=self.action_arc_phase, tar='muy_arc_from_cell',
                            value=self.muy_arc_target, tol=1e-8),
            ],
            vary=[
                xd.VaryList(self.mqt_knob_names, self.line.vars, step=1e-5),
            ])
        opt.solve()

        res = {kk: np.abs(self.line.vars[kk]._value) for kk in self.mqt_knob_names}

        # restore initial knob values
        if self.restore:
            for kk in self.mqt_knob_names:
                self.line.vars[kk] = mqt_knob_values[kk]
        return res

# For simplicity I use the same target for all arcs

def solve_optimizations(opts):
    for opt in opts:
        opt.solve()
        print(f'done optimization on {opt.vary[0].name}')

all_knobs = []
for arc in ['12', '23', '34', '45', '56', '67', '78', '81']:
    all_knobs += [f'kqf.a{arc}', f'kqd.a{arc}']
    all_knobs += [f'kqtf.a{arc}b1', f'kqtd.a{arc}b1']
    all_knobs += [f'kqtf.a{arc}b2', f'kqtd.a{arc}b2']

if __name__ == '__main__':

    # Load the line
    collider = xt.Environment.from_json(
        '../../test_data/hllhc15_collider/collider_00_from_mad.json')
    collider.build_trackers()

    collider.lhcb1.twiss_default['method'] = '4d'
    collider.lhcb2.twiss_default['method'] = '4d'
    collider.lhcb2.twiss_default['reverse'] = True

    twb1 = collider.lhcb1.twiss()
    mux_arc_target_b1 = twb1['mux', 's.ds.l7.b1'] - twb1['mux', 'e.ds.r6.b1']
    muy_arc_target_b1 = twb1['muy', 's.ds.l7.b1'] - twb1['muy', 'e.ds.r6.b1']

    twb2 = collider.lhcb2.twiss()
    mux_arc_target_b2 = twb2['mux', 's.ds.l7.b2'] - twb2['mux', 'e.ds.r6.b2']
    muy_arc_target_b2 = twb2['muy', 's.ds.l7.b2'] - twb2['muy', 'e.ds.r6.b2']

    optimizations_to_do =[]

    for ss in ['12', '23', '34', '45', '56', '67', '78', '81']:

        action_match_mqt_s67_b1 = ActionMatchPhaseWithMQTs(
            arc_name=ss, line_name='lhcb1', line=collider,
            mux_arc_target=mux_arc_target_b1, muy_arc_target=muy_arc_target_b1)
        action_match_mqt_s67_b2 = ActionMatchPhaseWithMQTs(
            arc_name=ss, line_name='lhcb2', line=collider,
            mux_arc_target=mux_arc_target_b2, muy_arc_target=muy_arc_target_b2)

        optimize_phase_arc = xd.Optimize(
            verbose=False,
            assert_within_tol=False,
            solver_options={'n_bisections': 3, 'min_step': 1e-5, 'n_steps_max': 5,},
            targets=[
                xd.Target(action=action_match_mqt_s67_b1, tar=f'kqtf.a{ss}b1', value=0),
                xd.Target(action=action_match_mqt_s67_b1, tar=f'kqtd.a{ss}b1', value=0),
                xd.Target(action=action_match_mqt_s67_b2, tar=f'kqtf.a{ss}b2', value=0),
                xd.Target(action=action_match_mqt_s67_b2, tar=f'kqtd.a{ss}b2', value=0),
            ],
            vary=[
                xd.Vary(name=f'kqf.a{ss}', container=collider.vars, step=1e-5),
                xd.Vary(name=f'kqd.a{ss}', container=collider.vars, step=1e-5),
        ])

        optimizations_to_do.append(optimize_phase_arc)

    optimizations_to_do = 8 * [optimizations_to_do[0]]

    initial_values = {kk: collider.vars[kk]._value for kk in all_knobs} # create setters

    # t1 = time.time()
    # pool = mp.Pool(processes=4)
    # pool.map(solve_optimizations, [optimizations_to_do[:2], optimizations_to_do[2:4],
    #                                 optimizations_to_do[4:6], optimizations_to_do[6:]])
    # t2 = time.time()
    # print(f'Time spent(parallel): {t2-t1} s')

    t1 = time.time()
    for ii, opt in enumerate(optimizations_to_do):
        # restore initial values
        for kk in all_knobs:
            collider.vars[kk] = initial_values[kk]
        print('pickling...')
        tp1 = time.perf_counter()
        opt = pickle.loads(pickle.dumps(opt))
        tp2 = time.perf_counter()
        print(f'pickling time: {tp2-tp1} s')
        print('solving...')
        ts1 = time.perf_counter()
        opt.solve()
        ts2 = time.perf_counter()
        print(f'Optimization {ii} done in {ts2-ts1} s')
    t2 = time.time()
    print(f'Time spent (serial): {t2-t1} s')

    # xt.general._print.suppress = False


import pathlib

import numpy as np
import pytest

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(__file__).parent.joinpath('../test_data').absolute()

@for_all_test_contexts
@pytest.mark.parametrize('freeze_longitudinal', [True, False])
def test_footprint(test_context, freeze_longitudinal):

    if isinstance(test_context, xo.ContextPyopencl):
        pytest.skip('Pyopencl not yet supported for footprint')
        return

    nemitt_x = 1e-6
    nemitt_y = 1e-6

    line = xt.Line.from_json(test_data_folder /
                'hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')
    line.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, p0c=7e12)
    line.build_tracker(_context=test_context)

    if freeze_longitudinal:
        kwargs = {'freeze_longitudinal': True}
        for ee in line.elements:
            if isinstance(ee, xt.Cavity):
                ee.voltage = 0
    else:
        kwargs = {}

    line.vars['i_oct_b1'] = 0
    fp0 = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                             **kwargs)

    line.vars['i_oct_b1'] = 500
    fp1 = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                            n_r=11, n_theta=7, r_range=[0.05, 7],
                            theta_range=[0.01, np.pi/2-0.01],
                            keep_tracking_data=True,
                            **kwargs)

    assert hasattr(fp1, 'tracking_data')

    assert hasattr(fp1, 'theta_grid')
    assert hasattr(fp1, 'r_grid')

    assert len(fp1.r_grid) == 11
    assert len(fp1.theta_grid) == 7

    xo.assert_allclose(fp1.r_grid[0], 0.05, rtol=0, atol=1e-10)
    xo.assert_allclose(fp1.r_grid[-1], 7, rtol=0, atol=1e-10)

    xo.assert_allclose(fp1.theta_grid[0], 0.01, rtol=0, atol=1e-10)
    xo.assert_allclose(fp1.theta_grid[-1], np.pi/2 - 0.01, rtol=0, atol=1e-10)

    #i_theta = 0, i_r = 0
    xo.assert_allclose(fp1.x_norm_2d[0, 0], 0.05, rtol=0, atol=1e-3)
    xo.assert_allclose(fp1.y_norm_2d[0, 0], 0, rtol=0, atol=1e-3)
    xo.assert_allclose(fp1.qx[0, 0], 0.31, rtol=0, atol=5e-5)
    xo.assert_allclose(fp1.qy[0, 0], 0.32, rtol=0, atol=5e-5)

    #i_theta = 0, i_r = 10
    xo.assert_allclose(fp1.x_norm_2d[0, -1], 7, rtol=0, atol=1e-3)
    xo.assert_allclose(fp1.y_norm_2d[0, -1], 0.07, rtol=0, atol=1e-3)
    xo.assert_allclose(fp1.qx[0, -1], 0.3129, rtol=0, atol=2e-4)
    xo.assert_allclose(fp1.qy[0, -1], 0.3185, rtol=0, atol=2e-4)

    #i_theta = 6, i_r = 0
    xo.assert_allclose(fp1.x_norm_2d[-1, 0], 0, rtol=0, atol=1e-3)
    xo.assert_allclose(fp1.y_norm_2d[-1, 0], 0.05, rtol=0, atol=1e-3)
    xo.assert_allclose(fp1.qx[0, 0], 0.31, rtol=0, atol=5e-5)
    xo.assert_allclose(fp1.qy[0, 0], 0.32, rtol=0, atol=5e-5)

    #i_theta = 6, i_r = 10
    xo.assert_allclose(fp1.x_norm_2d[-1, -1], 0.07, rtol=0, atol=1e-3)
    xo.assert_allclose(fp1.y_norm_2d[-1, -1], 7, rtol=0, atol=1e-3)
    xo.assert_allclose(fp1.qx[-1, -1], 0.3085, rtol=0, atol=2e-4)
    xo.assert_allclose(fp1.qy[-1, -1], 0.3229, rtol=0, atol=2e-4)

    xo.assert_allclose(np.max(fp1.qx[:]) - np.min(fp1.qx[:]), 4.4e-3, rtol=0, atol=1e-4)
    xo.assert_allclose(np.max(fp1.qy[:]) - np.min(fp1.qy[:]), 4.4e-3, rtol=0, atol=1e-4)

    line.vars['i_oct_b1'] = 0
    fp0 = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                             **kwargs)

    assert hasattr(fp0, 'theta_grid')
    assert hasattr(fp0, 'r_grid')

    xo.assert_allclose(fp0.r_grid[0], 0.1, rtol=0, atol=1e-10)
    xo.assert_allclose(fp0.r_grid[-1], 6, rtol=0, atol=1e-10)

    xo.assert_allclose(fp0.theta_grid[0], 0.05, rtol=0, atol=1e-10)
    xo.assert_allclose(fp0.theta_grid[-1], np.pi/2-0.05, rtol=0, atol=1e-10)

    assert len(fp0.r_grid) == 10
    assert len(fp0.theta_grid) == 10

    #i_theta = 0, i_r = 0
    xo.assert_allclose(fp0.x_norm_2d[0, 0], 0.1, rtol=0, atol=1e-3)
    xo.assert_allclose(fp0.y_norm_2d[0, 0], 0.005, rtol=0, atol=1e-3)
    xo.assert_allclose(fp0.qx[0, 0], 0.31, rtol=0, atol=5e-5)
    xo.assert_allclose(fp0.qy[0, 0], 0.32, rtol=0, atol=5e-5)

    xo.assert_allclose(np.max(fp0.qx[:]) - np.min(fp0.qx[:]), 0.0003, rtol=0, atol=2e-5)
    xo.assert_allclose(np.max(fp0.qy[:]) - np.min(fp0.qy[:]), 0.0003, rtol=0, atol=2e-5)

    line.vars['i_oct_b1'] = 500
    fp1_jgrid = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                                x_norm_range=[0.01, 6], y_norm_range=[0.01, 6],
                                n_x_norm=9, n_y_norm=8,
                                mode='uniform_action_grid',
                                **kwargs)

    assert hasattr(fp1_jgrid,  'Jx_grid')
    assert hasattr(fp1_jgrid,  'Jy_grid')

    assert len(fp1_jgrid.Jx_grid) == 9
    assert len(fp1_jgrid.Jy_grid) == 8

    xo.assert_allclose(np.diff(fp1_jgrid.Jx_grid), np.diff(fp1_jgrid.Jx_grid)[0],
                    rtol=0, atol=1e-10)
    xo.assert_allclose(np.diff(fp1_jgrid.Jy_grid), np.diff(fp1_jgrid.Jy_grid)[0],
                        rtol=0, atol=1e-10)

    xo.assert_allclose(fp1_jgrid.x_norm_2d[0, 0], 0.01, rtol=0, atol=1e-10)
    xo.assert_allclose(fp1_jgrid.y_norm_2d[0, 0], 0.01, rtol=0, atol=1e-10)
    xo.assert_allclose(fp1_jgrid.qx[0, 0], 0.31, rtol=0, atol=5e-5)
    xo.assert_allclose(fp1_jgrid.qy[0, 0], 0.32, rtol=0, atol=5e-5)

    xo.assert_allclose(fp1_jgrid.x_norm_2d[0, -1], 6, rtol=0, atol=1e-10)
    xo.assert_allclose(fp1_jgrid.y_norm_2d[0, -1], 0.01, rtol=0, atol=1e-10)
    xo.assert_allclose(fp1_jgrid.qx[0, -1], 0.3121, rtol=0, atol=1e-4)
    xo.assert_allclose(fp1_jgrid.qy[0, -1], 0.3189, rtol=0, atol=1e-4)

    xo.assert_allclose(fp1_jgrid.x_norm_2d[-1, 0], 0.01, rtol=0, atol=1e-10)
    xo.assert_allclose(fp1_jgrid.y_norm_2d[-1, 0], 6, rtol=0, atol=1e-10)
    xo.assert_allclose(fp1_jgrid.qx[-1, 0], 0.3089, rtol=0, atol=1e-4)
    xo.assert_allclose(fp1_jgrid.qy[-1, 0], 0.3221, rtol=0, atol=1e-4)

    xo.assert_allclose(fp1_jgrid.x_norm_2d[-1, -1], 6, rtol=0, atol=1e-10)
    xo.assert_allclose(fp1_jgrid.y_norm_2d[-1, -1], 6, rtol=0, atol=1e-10)
    xo.assert_allclose(fp1_jgrid.qx[-1, -1], 0.3111, rtol=0, atol=1e-4)
    xo.assert_allclose(fp1_jgrid.qy[-1, -1], 0.3210, rtol=0, atol=1e-4)

    xo.assert_allclose(np.max(fp1_jgrid.qx[:]) - np.min(fp1_jgrid.qx[:]), 0.0032,
                        rtol=0, atol=1e-4)
    xo.assert_allclose(np.max(fp1_jgrid.qy[:]) - np.min(fp1_jgrid.qy[:]), 0.0032,
                        rtol=0, atol=1e-4)

    x_norm_range = [1, 6]
    y_norm_range = [1, 6]
    line.vars['i_oct_b1'] = 50000 # Particles are lost for such high octupole current
    fp50k = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                                x_norm_range=x_norm_range, y_norm_range=y_norm_range,
                                n_x_norm=9, n_y_norm=8,
                                mode='uniform_action_grid',
                                linear_rescale_on_knobs=xt.LinearRescale(
                                knob_name='i_oct_b1', v0=500, dv=100),
                                **kwargs)


    line.vars['i_oct_b1'] = 60000 # Particles are lost for such high octupole current
    fp60k = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                                x_norm_range=x_norm_range, y_norm_range=y_norm_range,
                                n_x_norm=9, n_y_norm=8,
                                mode='uniform_action_grid',
                                linear_rescale_on_knobs=xt.LinearRescale(
                                knob_name='i_oct_b1', v0=500, dv=100),
                                **kwargs)

    line.vars['i_oct_b1'] = 500
    fp500 = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                                x_norm_range=x_norm_range, y_norm_range=y_norm_range,
                                n_x_norm=9, n_y_norm=8,
                                mode='uniform_action_grid',
                                **kwargs)

    line.vars['i_oct_b1'] = 600
    fp600 = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                                x_norm_range=x_norm_range, y_norm_range=y_norm_range,
                                n_x_norm=9, n_y_norm=8,
                                mode='uniform_action_grid',
                                **kwargs)

    xo.assert_allclose((fp60k.qx - fp50k.qx)/(fp600.qx-fp500.qx), 100, rtol=0, atol=1e-2)
    xo.assert_allclose((fp60k.qy - fp50k.qy)/(fp600.qy-fp500.qy), 100, rtol=0, atol=1e-2)

@for_all_test_contexts
def test_footprint_delta0(test_context):

    if isinstance(test_context, xo.ContextPyopencl):
        pytest.skip('Pyopencl not yet supported for footprint')
        return

    nemitt_x = 1e-6
    nemitt_y = 1e-6

    line = xt.Line.from_json(test_data_folder /
                        'hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')
    line.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, p0c=7e12)
    line.build_tracker(_context=test_context)

    # Some octupoles and chromaticity to see the footprint moving
    line.vars['i_oct_b1'] = 500
    line.match(
        targets=[xt.TargetList(['dqx', 'dqy'], value=10, tol=0.01)],
        vary=[xt.VaryList(['ksf.b1', 'ksd.b1'], step=1e-3)])

    # Compute and plot footprint on momentum
    fp1 = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                            freeze_longitudinal=True)

    # Compute and plot footprint off momentum
    fp2 = line.get_footprint(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                            freeze_longitudinal=True, delta0=1e-4)

    xo.assert_allclose(line.record_last_track.delta[:], 1e-4, atol=1e-12, rtol=0)

    # It is 12, 9 instead of 10, 10 because of non-linear chromaticity
    xo.assert_allclose((np.max(fp2.qx) - np.max(fp1.qx))/1e-4, 12, atol=0.5, rtol=0)
    xo.assert_allclose((np.max(fp2.qy) - np.max(fp1.qy))/1e-4, 9, atol=0.5, rtol=0)




import numpy as np
import xtrack as xt

# Load a line and build tracker
# line = xt.Line.from_json(
#     '../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line = xt.Line.from_json(
    '../../test_data/hllhc14_no_errors_with_coupling_knobs/line_b1.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.cycle('ip1', inplace=True)
line.twiss_default['method'] = '4d'

# Flat machine and skew quadrupoles off
for nn in line.vars.get_table().rows['on_.*|acb.*|cm.*'].name:
    line.vars[nn] = 0

# see Eq. 47 in https://cds.cern.ch/record/522049/files/lhc-project-report-501.pdf
class ActionCmin(xt.Action):
    def __init__(self, line):
        self.line = line
    def run(self):
        tw = self.line.twiss()
        tt = self.line.get_table(attr=True)
        k1sl = tt['k1sl']
        c_min = 1 / (2*np.pi) * np.sum(k1sl * np.sqrt(tw.betx * tw.bety)
                                * np.exp(1j * 2 * np.pi * (tw.mux - tw.muy)))
        return {'c_min_re': c_min.real, 'c_min_im': c_min.imag}

act_cmin = ActionCmin(line)

# vary=[xt.VaryList(['kqs.a23b1', 'kqs.a45b1', 'kqs.a67b1', 'kqs.a81b1'], step=5e-5),
#       xt.VaryList(['kqs.l2b1','kqs.l4b1', 'kqs.l6b1', 'kqs.l8b1',
#                    'kqs.r1b1', 'kqs.r3b1', 'kqs.r5b1', 'kqs.r7b1'],
#                    step=5e-5)]


# vary=[xt.VaryList(['kqs.a23b1', 'kqs.a67b1'], step=5e-5),
#       xt.VaryList(['kqs.l4b1', 'kqs.l8b1','kqs.r3b1', 'kqs.r7b1'], weight=2,
#                   step=5e-5)]

line.vars['kqs.a34b1'] = 0
line.vars['kqs.a78b1'] = 0
line.vars['kqs.l4b1'] = line.vars['kqs.a34b1']
line.vars['kqs.l8b1'] = line.vars['kqs.a78b1']
line.vars['kqs.r3b1'] = line.vars['kqs.a34b1']
line.vars['kqs.r7b1'] = line.vars['kqs.a78b1']
vary=xt.VaryList(['kqs.a23b1', 'kqs.a34b1', 'kqs.a67b1', 'kqs.a78b1'], step=5e-5)


c_min_match = 1e-4

opt_re = line.match_knob(knob_name='c_minus_re.b1',
    knob_value_start=0, knob_value_end=c_min_match,
    run=False,
    vary=vary,
    targets=[
        act_cmin.target('c_min_re', value=c_min_match, tol=1e-8),
        act_cmin.target('c_min_im', value=0, tol=1e-8),
    ])
opt_re.solve()
opt_re.generate_knob()

opt_im = line.match_knob(knob_name='c_minus_im.b1',
    knob_value_start=0, knob_value_end=c_min_match,
    run=False,
    vary=vary,
    targets=[
        act_cmin.target('c_min_re', value=0, tol=1e-8),
        act_cmin.target('c_min_im', value=c_min_match, tol=1e-8),
    ])
opt_im.solve()
opt_im.generate_knob()

line.vars['c_minus_re.b1'] = 5e-3
line.vars['c_minus_im.b1'] = 0
tt_re = line.get_table(attr=True)

line.vars['c_minus_re.b1'] = 0
line.vars['c_minus_im.b1'] = 5e-3
tt_im = line.get_table(attr=True)

# Check orthogonality
line.vars['c_minus_re.b1'] = 1e-3
line.vars['c_minus_im.b1'] = 1e-3
assert np.isclose(line.twiss().c_minus/np.sqrt(2), 1e-3, rtol=0, atol=1.5e-5)

# Compare against "legacy" knobs
line_legacy = xt.Line.from_json(
    '../../test_data/hllhc14_no_errors_with_coupling_knobs/line_b1.json')
line_legacy.cycle('ip1', inplace=True)

# Flat machine and skew quadrupoles off
for nn in line_legacy.vars.get_table().rows['on_.*|acb.*|cm.*'].name:
    line_legacy.vars[nn] = 0

tw_leg = line_legacy.twiss()

line_legacy.vars['cmrskew'] = 5e-3
line_legacy.vars['cmiskew'] = 0
tt_re_leg = line_legacy.get_table(attr=True)

line_legacy.vars['cmrskew'] = 0
line_legacy.vars['cmiskew'] = 5e-3
tt_im_leg = line_legacy.get_table(attr=True)

# Withing 12% of the maximum value
assert np.allclose(tt_re.k1sl, tt_re_leg.k1sl, rtol=0,
                   atol=0.12 * np.max(tt_re_leg.k1sl))
assert np.allclose(tt_im.k1sl, tt_im_leg.k1sl, rtol=0,
                   atol=0.12 * np.max(tt_im_leg.k1sl))

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
ax1 = plt.subplot(1, 1, 1)
plt.plot(tw_leg.s, tt_re_leg['k1sl'], '.', label='cmrskew=1e-3')
plt.plot(tw_leg.s, tt_im_leg['k1sl'], '.', label='cmiskew=1e-3')
plt.ylabel('k1s * L')
plt.xlabel('s [m]')
plt.suptitle('Legacy knobs')
plt.legend()

plt.figure(2)
ax21 = plt.subplot(1, 1, 1, sharey=ax1)
plt.plot(tt_re.s, tt_re['k1sl'], '.', label='c_minus_re.b1=1e-3')
plt.plot(tt_im.s, tt_im['k1sl'], '.', label='c_minus_im.b1=1e-3')
plt.ylabel('k1s * L')
plt.xlabel('s [m]')
plt.suptitle('Xsuite')
plt.legend()

plt.show()



